/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

require([
    "app/common",
    "app/quotas",
    "app/messages",
    "aps/load",
    "aps/ResourceStore",
    "aps/Memory",
    "aps/xhr",
    "dojo/promise/all",
    "dojox/mvc/at",
    "dojox/mvc/getStateful",
    "dojox/mvc/getPlainValue",
    "dijit/registry"
], function (
    common,
    quotas,
    messages,
    load,
    Store,
    Memory,
    xhr,
    all,
    at,
    getStateful,
    getPlainValue,
    registry
) {
    var tenant = aps.context.vars.tenant,
        tenantId = tenant.aps.id,
        account = aps.context.vars.user,
        accountStore = new Store({
            apsType: common.TYPES.USERS,
            target: "/aps/2/resources"
        }),
        quotasStatisticsDefer = xhr.get("/aps/2/resources/" + tenantId + "/quotasStatistics"),
        backupProfileDefer = xhr.get("/aps/2/resources/" + tenantId + "/profile"),
        controls = null;

    function renderQuota(quotasStatistics, userQuota) {
        var quotaContainer = registry.byId("quotaContainer"),
            offeringItemsToDisplay = common.getAvailableCountersForArea('quota_controls_ccpv1_edit_user').filter(function(counter){
                return counter.name.indexOf('Secondary') < 0;
            }),
            quotaIsOn = userQuota != common.USER_QUOTA.OFF;
        controls = quotas.getQuotaControls(
            account, tenant, offeringItemsToDisplay, quotasStatistics, false, userQuota
        );

        Object.entries(controls).forEach(function ([counterName, control]) {
            var userOfferingItem = common.getUserOfferingItem(account, counterName, true);
            if (userOfferingItem && (quotaIsOn || userOfferingItem.limit > 0)) {
                var counterData = common.getCountersToOfferingItems()[common.getStandardCounterName(counterName)];
                control.set("value", userOfferingItem.limit == common.QUOTA_UNLIMITED_VALUE ?
                    "" : (counterData.legendSchema === 'storage' ?
                        userOfferingItem.limit / common.BYTES_IN_GB : userOfferingItem.limit));
                quotaContainer.addChild(control);
            }
        });
    }

    function applyQuotaValues(userData, userQuota) {
        let getLimitValue = function (limitFieldValue, isStorage) {
            return limitFieldValue === "" ? common.QUOTA_UNLIMITED_VALUE
                : limitFieldValue * (isStorage ? common.BYTES_IN_GB : 1);
        };
        if (controls) {
            Object.entries(controls).forEach(function ([counterName, control]) {
                let offeringItem = common.getUserOfferingItem(userData, counterName, true);
                if (offeringItem === null) {
                    return;
                }
                offeringItem.limit = control.value !== ''
                    ? getLimitValue(control.value, common.isStorage(counterName))
                    : userQuota === common.USER_QUOTA.STRICT ? 0 : -1;
            });
        }
    }

    all({
        profiles: backupProfileDefer,
        quotasStatistics: quotasStatisticsDefer,
        apsUser: common.fetchTenantUser(tenantId, account.aps.id) // workaround for not loaded users
    }).then(function (response) {
        account = account.offeringItems ? account : response.apsUser;
        var backupProfile = response.profiles.length ? response.profiles[0] : null,
            quotasStatistics = response.quotasStatistics,
            userQuota = backupProfile ? backupProfile.userQuota : common.USER_QUOTA.STRICT,
            showQuotas = quotas.canShowQuotas(userQuota, account),
            quotaIsOn = userQuota != common.USER_QUOTA.OFF,
            userModel = getStateful(account),
            privilegedCheckBoxState = getStateful({}),
            accountStatus;

        userModel.notifications = common.transformNotificationsToObject(userModel.notifications);

        accountStatus = userModel.userstatus == common.USER_STATUS.NOT_ACTIVATED ?
            [["aps/Output", {
                label: _("Status"),
                value:_("Not activated")
            }]] :
            [["aps/RadioButton", {
                id: "enableButton",
                label: _('Status'),
                name: "status",
                checked: at(userModel, "servicestatus").transform({
                    format: function (value) {
                        return parseInt(value, 10) === 1;
                    },
                    parse: function (value) {
                        return value ? 1 : 0;
                    }
                }),
                description: _('Enabled')
            }],
                ["aps/RadioButton", {
                    id: "disableButton",
                    name: "status",
                    checked: at(userModel, "servicestatus").transform({
                        format: function (value) {
                            return parseInt(value, 10) !== 1;
                        },
                        parse: function (value) {
                            return value ? 0 : 1;
                        }
                    }),
                    description: _('Disabled'),
                    hint: _('Disabling an account makes backup and recovery operations under this account unavailable. The backed-up data is kept until you delete the account.')
                }]];

		load(["aps/PageContainer", {id: "apsPageContainer"}, [
			["aps/FieldSet", {
				title: _("Main")
			}, [
				["aps/Output", {
					label: _("User name"),
					value: common.escapeAngularBrackets(userModel.displayName + ' (' + userModel.endUserLogin + ')')
				}],
				["aps/CheckBox", {
					id: "privileged",
					label: _("Administrator privileges"),
					checked: at(userModel, "privileged").transform({
						format: function (value) {
							return parseInt(value, 10) === common.USER_PRIVILEGED.YES;
						},
						parse: function (value) {
							return (value) ? common.USER_PRIVILEGED.YES : common.USER_PRIVILEGED.NO;
						}
					}),
					disabled: at(privilegedCheckBoxState, "disabled")
				}]
			]],
			["aps/FieldSet", {
				escapeHTML: false,
				title: _('Account status')
			}, accountStatus
			],
			["aps/FieldSet", {
				title: _('Notifications')
			}, [
				["aps/FieldSet", {
					label: _('MANAGEMENT PORTAL'),
					title: false
				}],
				["aps/CheckBox", {
					label: _('Quota overuse notifications'),
					checked: at(userModel.notifications, "quota")
				}],
				["aps/CheckBox", {
					label: _('Scheduled usage reports'),
					checked: at(userModel.notifications, "reports")
				}],
				["aps/Output", {
					class: 'space-before-element'
				}],
				["aps/FieldSet", {
					label: _('CYBER PROTECTION'),
					title: false
				}],
				["aps/CheckBox", {
					label: _('Failure notifications'),
					checked: at(userModel.notifications, "backup_error")
				}],
				["aps/CheckBox", {
					label: _('Warning notifications'),
					checked: at(userModel.notifications, "backup_warning")
				}],
				["aps/CheckBox", {
					label: _('Success notifications'),
					checked: at(userModel.notifications, "backup_info")
				}],
				["aps/CheckBox", {
					label: _('Daily recap about active alerts'),
					checked: at(userModel.notifications, "backup_daily_report")
				}],
				["aps/CheckBox", {
					label: _('Malware detection notifications'),
					checked: at(userModel.notifications, "anti_malware")
				}]
			]],
            ["aps/Output", {
                class: 'space-before-element-block'
            }],
            ["aps/FieldSet", {
                id: "quotaContainer",
                visible: quotaIsOn,
                description: quotas.getQuotaDescription(userQuota),
                title: _('Quotas'),
                disabled: at(userModel, "servicestatus").transform({
                    format: function (value) {
                        return parseInt(value, 10) !== 1;
                    }
                }).direction(at.from)
            }]
        ]]).then(function () {
            if (quotaIsOn) {
                renderQuota(quotasStatistics, userQuota);
            } else if (showQuotas) {
                messages.displayWarning(_("After submitting this form, all user quotas will become unlimited."));
            }

            function setPrivilegedCheckboxState(name, oldValue, newValue) {
                if (typeof newValue == "undefined") {
                    return;
                }
                privilegedCheckBoxState.set("disabled", parseInt(newValue, 10) !== 1);
            }

            userModel.watch("servicestatus", setPrivilegedCheckboxState);
            privilegedCheckBoxState.set("disabled", parseInt(userModel.servicestatus, 10) !== 1);

            aps.app.onSubmit = function () {
                var page = registry.byId("apsPageContainer");
                if (!page.validate()) {
                    aps.apsc.cancelProcessing();
                    return;
                }
                if (quotaIsOn) {
                    applyQuotaValues(userModel, userQuota);
                }

                if (Object.keys(userModel.notifications).length) {
                    userModel.notifications = common.transformNotificationsToArray(userModel.notifications);
                } else {
                    delete userModel.notifications;
                }
                accountStore.put(getPlainValue(userModel)).then(function () {
                    aps.apsc.gotoView(common.getRelatedViewId('list'));
                }, messages.displayError);
            };
        });
    });

    aps.app.onCancel = function () {
        aps.apsc.gotoView(common.getRelatedViewId('list'));
    };
});



