/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

define([
    "dojo/Deferred",
    "dojo/promise/all",
    "dojox/mvc/at",
    "dojox/mvc/getStateful",
    "aps/xhr",
    "aps/ResourceStore",
    "aps/Memory",
    "aps/FieldSet",
    "aps/Grid",
    "aps/Panel",
    "aps/Output",
    "aps/CheckBox",
    "aps/Container",
    "aps/Spinner",
    "./getStorageOutput",
    "./getResourceOutput",
    "./../../common",
    "./../../uomConverter"
], function (
    Deferred,
    all,
    at,
    getStateful,
    xhr,
    Store,
    Memory,
    FieldSet,
    Grid,
    Panel,
    Output,
    CheckBox,
    Container,
    Spinner,
    getStorageOutput,
    getResourceOutput,
    common,
    uomConverter
) {
    function convert(value, from, to) {
        if (from === "unit") {
            return value;
        }

        if (value > 0) {
            return uomConverter.convert(value, from, to);
        }
        return value;
    }

    function getBillingResources(tenant) {
        return aps.biz.getResourcesInfo({
            filter: [{apsType: common.TYPES.TENANT}],
            subscriptionId: tenant.aps.subscription
        });
    }

    function findBillingResource(resource, billingResources) {
        var result = null;
        billingResources.some(function (billingResource) {
            if (resource.apsType === billingResource.apsType && resource.title === billingResource.name) {
                result = billingResource;
                return true;
            }
            return false;
        });
        return result;
    }

    /**
     * Returns array of global variables
     * @param tenant
     * @returns Deferred
     */
    function getGlobalVariables(tenant) {
        var deferred = new Deferred();
        xhr.get('/aps/2/resources/' + tenant.aps.id + '/global').then(function (response) {
            deferred.resolve(response[0]);
        }, deferred.reject);

        return deferred;
    }

    /**
     * Returns profile
     * @param tenant
     * @returns Deferred
     */
    function getProfile(tenant) {
        var deferred = new Deferred();
        xhr.get('/aps/2/resources/' + tenant.aps.id + '/profile').then(function (response) {
            deferred.resolve(response[0]);
        }, deferred.reject);

        return deferred;
    }

    function sumBackupAccountsCounters(usersUsages) {
        function addResource(counter, resource) {
            if (!resource) {
                return;
            }

            counter.usage = addUsage(counter.usage, resource.usage);
            counter.allocated = addAllocated(counter.allocated, resource.limit);
        }

        function addUsage(usage, value) {
            if (!value) {
                return usage;
            }
            return (usage + value);
        }

        function addAllocated(allocated, value) {
            if (!value) {
                return allocated;
            }
            if (allocated < 0 || value < 0) {
                return -1;
            }
            return (allocated + value);
        }

        let countersList = common.getAvailableCountersForArea('user_quotas');
        let counters = {};
        for (let userId in usersUsages) {
            let userUsage = usersUsages[userId];
            countersList.forEach(function (counter) {
                if (!userUsage[counter.name]) {
                    return;
                }
                counters[counter.name] = counters[counter.name] || {usage: 0, allocated: 0};
                addResource(counters[counter.name], userUsage[counter.name]);
            });
        }

        return counters;
    }

    function getBackupAccounts(tenant) {
        return xhr.get('/aps/2/resources/' + tenant.aps.id + '/updateStatistics');
    }

    function getSubscription(tenant) {
        var deferred = new Deferred();
        if (tenant.subscription) {
            deferred.resolve(tenant.subscription);
        } else {
            var store = new Store({target: "/aps/2/resources/" + tenant.aps.id + "/subscription"});
            store.query().then(function (subscriptions) {
                subscriptions && subscriptions.length ? deferred.resolve(subscriptions[0]) : deferred.resolve(null);
            }, deferred.reject);
        }
        return deferred;
    }

    function setDeltas(deltas, delta, apsType, apsProperty) {
        var deltaObj = deltas.find(function (item) {
            return item.apsType === apsType && item.apsProperty === apsProperty;
        });

        if (deltaObj) {
            deltaObj.delta = delta;
            return;
        }

        deltas.push({apsType: apsType, apsProperty: apsProperty, delta: delta});
    }

    function getUserResourcesGrid(tenant, model, divisor) {
        var showUnlimitedColumn = false,
            newUser = !(model.aps && model.aps.id),
            visibleCounters = common.getTenantVisibleQuotas(tenant, "user_quotas"),
            gridState = getStateful(Object.values(visibleCounters).reduce(
                function (grid, counterName) {
                    grid.unlimitedValue[counterName] = false;
                    grid.spinnerValue[counterName] = 0;
                    grid.counterDisabled[counterName] = false;

                    return grid;
                },
                {
                    unlimitedValue: {},
                    spinnerValue: {},
                    counterDisabled: {}
                }
            )),
            deltas = [];

        var deferred = new Deferred();
        all({
            subscription: getSubscription(tenant),
            subscriptionResources: common.getBackupAccountResources(tenant),
            billingResources: getBillingResources(tenant),
            backupAccounts: getBackupAccounts(tenant),
            global: getGlobalVariables(tenant),
            profile: getProfile(tenant)
        }).then(function (response) {
            var accounts = sumBackupAccountsCounters(response.backupAccounts.userUsage),
                resources = [];
            response.subscriptionResources.forEach(function (subscriptionResource) {
                var resourceId = subscriptionResource.id,
                    resourceName = subscriptionResource.title,
                    apsType = subscriptionResource.apsType,
                    apsProperty = subscriptionResource.property,
                    resourceUnit = subscriptionResource.unit,
                    tenantLimit = tenant[apsProperty].limit,
                    tenantUsage = tenant[apsProperty].usage ? tenant[apsProperty].usage : 0,
                    allocated = accounts[apsProperty] ? accounts[apsProperty].allocated : 0,
                    offeringItem = common.getUserOfferingItem(model, apsProperty),
                    userUsage = model.userUsage[apsProperty] && model.userUsage[apsProperty].usage ? model.userUsage[apsProperty].usage : 0,
                    userLimit = offeringItem && offeringItem.limit ? offeringItem.limit : 0,
                    billingResource = findBillingResource(subscriptionResource, response.billingResources),
                    upSell = !!billingResource && !response.subscription.trial && tenantLimit >= 0;

                if (apsProperty.indexOf('Secondary') > -1) {
                    return;
                }

                if (!tenantLimit && !upSell) {
                    return;
                }

                var available;
                if (tenantLimit < 0) {
                    available = -1;
                    showUnlimitedColumn = true;
                } else if (allocated < 0) {
                    available = 0;
                } else {
                    available = tenantLimit - allocated;
                    if (available < 0) {
                        available = 0;
                    }
                }

                var unit = ['unit', 'unith'].includes(resourceUnit) ? "unit" : "b",
                    spinnerUnit = ['unit', 'unith'].includes(resourceUnit) ? "unit" : "gb",
                    spinnerLegend = ['unit', 'unith'].includes(resourceUnit) ? "" : _("GB"),
                    resource = {
                        id: resourceId,
                        name: resourceName,

                        unit: unit,
                        spinnerUnit: spinnerUnit,
                        spinnerLegend: spinnerLegend,

                        apsType: apsType,
                        apsProperty: apsProperty,

                        userUsage: userUsage,
                        userLimit: userLimit,

                        tenantLimit: Number(convert(tenantLimit, unit, spinnerUnit).toFixed(0)),
                        tenantUsage: Number(Math.ceil(convert(tenantUsage, unit, spinnerUnit)).toFixed(0)),
                        allocated: Number(Math.floor(convert(allocated, unit, spinnerUnit)).toFixed(0)),
                        available: Number(convert(available, unit, spinnerUnit).toFixed(0)),
                        upSell: upSell,
                        upSellMax: billingResource !== null && typeof billingResource.max !== "undefined" ? billingResource.max : null,
                        price: upSell ? billingResource.priceText : null
                    };
                resources.push(resource);
            });
            resources.forEach(function (resource) {
                var spinnerValue, modelValue,
                    offeringItem = common.getUserOfferingItem(model, resource.apsProperty, true);
                if (newUser) {
                    if (resource.tenantLimit < 0) {
                        gridState.unlimitedValue[resource.apsProperty] = true;
                        modelValue = -1;
                        spinnerValue = 0;
                    } else {
                        gridState.unlimitedValue[resource.apsProperty] = false;
                        modelValue = spinnerValue = Number(Math.floor(resource.available / divisor));
                    }
                } else {
                    if (resource.userLimit < 0) {
                        if (resource.available == -1) {
                            gridState.unlimitedValue[resource.apsProperty] = true;
                            modelValue = -1;
                        } else {
                            gridState.unlimitedValue[resource.apsProperty] = false;
                            modelValue = 0;
                        }
                        spinnerValue = 0;
                    } else {
                        gridState.unlimitedValue[resource.apsProperty] = false;
                        modelValue = spinnerValue = Number(convert(resource.userLimit, resource.unit, resource.spinnerUnit).toFixed(0));
                    }
                }
                gridState.spinnerValue[resource.apsProperty] = spinnerValue;
                offeringItem.set("limit", Number(convert(modelValue, resource.spinnerUnit, resource.unit).toFixed(0)));
            });
            resources.sort(function (a, b) {
                return Number(a.id) - Number(b.id);
            });

            var store = new Memory({data: resources, idProperty: "id"});
            var grid = new Grid({
                noDataText: '',
                showPaging: false,
                pageSizeOptions: "GRID_UNLIMITED",
                store: store,
                columns: [{
                    "name": _("Quota"),
                    "sortable": false,
                    "field": "name",
                    renderCell: function (raw) {
                        var fs = new FieldSet();
                        fs.addChild(new Output({value: _(raw.name)}));
                        if (raw.upSell) {
                            fs.addChild(new Output({value: raw.price, alertType: "info"}));
                        }
                        return fs;
                    }
                }, {
                    "name": _("Current User's Usage"),
                    "sortable": false,
                    "field": "usage",
                    "visible": !newUser,
                    renderCell: function (raw) {
                        if (raw.unit === "unit") {
                            return getResourceOutput({
                                noneLegend: _("None"),
                                limitedLegend: _("${usage} of ${limit}"),
                                zeroUsageLegend: _("${limit} available"),
                                unlimitedLegend: _("${usage}"),
                                usage: raw.userUsage || 0,
                                limit: raw.userLimit || 0
                            });
                        }

                        return getStorageOutput({
                            noneLegend: _("None"),
                            limitedLegend: _("${usage} ${usageUnit} of ${limit} ${limitUnit}"),
                            unlimitedLegend: _("${usage} ${usageUnit}"),
                            usage: raw.userUsage || 0,
                            limit: raw.userLimit || 0
                        });
                    }
                }, {
                    "name": _("Available for Assigning"),
                    "sortable": false,
                    renderCell: function (raw) {
                        if (raw.tenantLimit < 0) {
                            return new Output({
                                value: _("Unlimited")
                            });
                        }

                        if (raw.unit === "unit") {
                            return new Output({
                                content: (_('${available} of ${limit}')),
                                limit: raw.tenantLimit,
                                available: raw.available
                            });
                        }

                        return new Output({
                            content: (_('${available} ${availableUnit} of ${limit} ${limitUnit}')),
                            limit: raw.tenantLimit,
                            limitUnit: _("GB"),
                            available: raw.available,
                            availableUnit: _("GB")
                        });
                    }
                }, {
                    "name": _("Unlimited"),
                    "sortable": false,
                    "align": "center",
                    visible: showUnlimitedColumn,
                    renderCell: function (raw) {
                        var offeringItem = common.getUserOfferingItem(model, raw.apsProperty);
                        return new CheckBox({
                            checked: at(gridState.unlimitedValue, raw.apsProperty).transform({
                                parse: function (value) {
                                    if (value) {
                                        offeringItem.set("limit", -1);
                                    } else {
                                        offeringItem.set("limit", Number(convert(Number(gridState.spinnerValue[raw.apsProperty]), raw.spinnerUnit, raw.unit).toFixed(0)));
                                    }
                                    return value;
                                }
                            }),
                            disabled: raw.tenantLimit >= 0 || at(gridState.counterDisabled, raw.apsProperty)
                        });
                    }
                }, {
                    "name": _("New Quota"),
                    "sortable": false,
                    "align": "left",
                    renderCell: function (raw) {
                        var userLimit = Number(convert(raw.userLimit, raw.unit, raw.spinnerUnit).toFixed(0)),
                            available = raw.available + (userLimit < 0 ? 0 : userLimit),
                            maximum = (raw.available < 0) ? null : Number(Math.floor(available / divisor)),
                            offeringItem = common.getUserOfferingItem(model, raw.apsProperty);

                        var spinner = new Spinner({
                            value: at(gridState.spinnerValue, raw.apsProperty).transform({
                                parse: function (value) {
                                    offeringItem.set("limit", Number(convert(Number(value), raw.spinnerUnit, raw.unit).toFixed(0)));

                                    if (raw.upSell) {
                                        var increase = (value * divisor) > available ? ((value * divisor) - available) : 0,
                                            delta = (raw.tenantLimit - raw.tenantUsage) + increase; // workaround!

                                        setDeltas(deltas, delta, raw.apsType, raw.apsProperty);
                                        aps.biz.requestUsageChange({
                                            deltas: deltas,
                                            operations: [],
                                            subscriptionId: tenant.aps.subscription
                                        });
                                    }
                                    return value;
                                }
                            }),
                            minimum: 0,
                            maximum: raw.upSell ? raw.upSellMax : maximum,
                            step: 1,
                            legend: raw.spinnerLegend,
                            invalidMessage: _("This field can only contain numbers. No spaces or other characters allowed."),
                            extraValidator: function (value) {
                                return Number.isInteger(+value);
                            }
                        });

                        function disableSpinner() {
                            spinner.set("disabled", (gridState.unlimitedValue[raw.apsProperty] || gridState.counterDisabled[raw.apsProperty]));
                        }

                        disableSpinner();
                        gridState.unlimitedValue.watch(raw.apsProperty, disableSpinner);
                        gridState.counterDisabled.watch(raw.apsProperty, disableSpinner);

                        return spinner;
                    }
                }]
            });

            var wrapPanel = new Panel({'title': _("User quotas")});
            wrapPanel.addChild(grid);
            deferred.resolve({
                grid: wrapPanel,
                deltas: deltas
            });
        }, deferred.reject);
        return deferred;
    }

    return getUserResourcesGrid;
});