/*
 * @Copyright © 2002-2021 Acronis International GmbH. All rights reserved
 */

define([
    "aps/Memory",
    "aps/_View",
    "aps/xhr",
    "aps/FieldSet",
    "dojo/_base/declare",
    "dijit/registry",
    "dojox/mvc/at",
    "dojox/mvc/getPlainValue",
    "dojox/mvc/getStateful",
    "./../../../common",
    "./../../../messages"
], function (
    Memory,
    _View,
    xhr,
    FieldSet,
    declare,
    registry,
    at,
    getPlainValue,
    getStateful,
    common,
    messages
) {
    return declare(_View, {

        _adminModel: null,
        _tenant: null,
        _showDatacenters: false,
        _datacenters: [],

        init: function () {
            var self = this;
            self._adminModel = getStateful({
                notifications: common.transformNotificationsToObject(common.RESELLER_NOTIFICATIONS)
            });
            self._datacenterModel = getStateful({id: null});
            self._formModel = getStateful({
                showDatacenters: false,
                datacenters: new Memory({data: [], idProperty: "value"})
            });

            return [
                ["aps/Panel", [
                    ["aps/FieldSet", {title: _("General information")}, [
                        ["aps/FieldSet", [
                            ["aps/TextBox", {
                                id: self.genId('login'),
                                label: _("Login"),
                                required: true,
                                trim: true,
                                value: at(self._adminModel, "login")
                            }],
                            ["aps/TextBox", {
                                id: self.genId('email'),
                                label: _("Email"),
                                required: true,
                                trim: true,
                                pattern: '^[^\\s@]+@[^\\s@]+\\.[^\\s@.]+$',
                                invalidMessage: _('Please enter a valid email.'),
                                value: at(self._adminModel, "email")
                            }]
                        ]],
                        ["aps/FieldSet", [
                            ["aps/Select", {
                                id: "datacenterId",
                                label: _("Datacenter"),
                                required: true,
                                gridSize: 'md-4',
                                visible: at(self._formModel, 'showDatacenters'),
                                store: at(self._formModel, 'datacenters'),
                                value: at(self._datacenterModel, 'id')
                            }]
                        ]]
                    ]],
                    ["aps/FieldSet", {title: _("Optional")}, [
                        ["aps/FieldSet", [
                            ["aps/TextBox", {
                                id: self.genId("first_name"),
                                label: _("First name"),
                                required: false,
                                optionalText: "",
                                trim: true,
                                value: at(self._adminModel, "firstName")
                            }],
                            ["aps/TextBox", {
                                id: self.genId("last_name"),
                                label: _("Last name"),
                                required: false,
                                optionalText: "",
                                trim: true,
                                value: at(self._adminModel, "lastName")
                            }]
                        ]]
                    ]],
                    ["aps/FieldSet", {"id": common.genId('fieldset'), title: _("Notifications")}, [
                        ["aps/FieldSet", {
                            label: _('MANAGEMENT PORTAL')
                        }, [
                            ["aps/CheckBox", {
                                label: _('Quota overuse notifications'),
                                checked: at(self._adminModel.notifications, "quota"),
                                gridSize: 'md-4'
                            }],
                            ["aps/CheckBox", {
                                label: _('Scheduled usage reports'),
                                checked: at(self._adminModel.notifications, "reports"),
                                gridSize: 'md-4'
                            }]
                        ]],
                        ["aps/FieldSet", {
                            label: _('CYBER PROTECTION')
                        }, [
                            ["aps/CheckBox", {
                                label: _('Failure notifications'),
                                checked: at(self._adminModel.notifications, "backup_error"),
                                gridSize: 'md-4'
                            }],
                            ["aps/CheckBox", {
                                label: _('Warning notifications'),
                                checked: at(self._adminModel.notifications, "backup_warning"),
                                gridSize: 'md-4'
                            }],
                            ["aps/CheckBox", {
                                label: _('Success notifications'),
                                checked: at(self._adminModel.notifications, "backup_info"),
                                gridSize: 'md-4'
                            }],
                            ["aps/CheckBox", {
                                label: _('Daily recap about active alerts'),
                                checked: at(self._adminModel.notifications, "backup_daily_report"),
                                gridSize: 'md-4'
                            }],
                            ["aps/CheckBox", {
                                label: _('Malware detection notifications'),
                                checked: at(self._adminModel.notifications, "anti_malware"),
                                gridSize: 'md-4'
                            }]
                        ]]
                    ]]
                ]]
            ];
        },
        onHide: function () {
            messages.removeAll();
        },
        onContext: function (context) {
            var self = this;
            self._tenant = context.vars.tenant;
            self._adminModel.set({
                login: "",
                email: "",
                firstName: "",
                lastName: "",
                notifications: common.transformNotificationsToObject(common.RESELLER_NOTIFICATIONS)
            });
            xhr.get("/aps/2/resources/" + self._tenant.aps.id + "/getDatacenters").then(function (datacenters) {
                var showDatacenters = datacenters.length !== 1;
                self._formModel.set('showDatacenters', showDatacenters);
                self._datacenterModel.set('id', datacenters.length ? datacenters[0].id : null);
                self._formModel.set('datacenters', new Memory({
                    data: datacenters.map(function (datacenter) {
                        return {
                            value: datacenter.id,
                            label: datacenter.name
                        };
                    }), idProperty: "value"
                }));
            }, messages.displayError);
            aps.apsc.hideLoading();
        },
        onSubmit: function () {
            var self = this,
                tenant = self._tenant,
                adminModel = self._adminModel,
                datacenterModel = self._datacenterModel;

            if (!self.validate()) {
                aps.apsc.cancelProcessing();
                return;
            }
            xhr.get("/aps/2/resources/" + tenant.aps.id + "/loginExists", {
                query: {
                    login: adminModel.login,
                    datacenterId: datacenterModel.id
                }
            }).then(function (existing) {
                if (existing) {
                    messages.displayWarning(_("This login already exists."));
                    aps.apsc.cancelProcessing();
                    return;
                }
                if (Object.keys(adminModel.notifications).length) {
                    adminModel.notifications = common.transformNotificationsToArray(adminModel.notifications);
                } else {
                    delete adminModel.notifications;
                }
                xhr("/aps/2/resources/" + tenant.aps.id + "/administrators/", {
                    method: "POST",
                    query: {
                        datacenterId: datacenterModel.id
                    },
                    data: JSON.stringify(getPlainValue(adminModel)),
                    headers: {"Content-Type": "application/json"}
                }).then(function () {
                    aps.apsc.gotoView("nextcp-ccp-reseller");
                }, messages.displayError);
            }, messages.displayError);
        }
    });
});
