<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Base;

use Acronis\App\Base\ApplicationCounter as Counter;
use Acronis\App\CloudApi\Api;
use Arr;

trait LegacyOfferingItemsTrait
{
    /**
     * @param array $limits
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function addMailboxInstanceSizeCounter(array &$limits)
    {
        // do nothing if the counter is already specified in the service plan
        if (Arr::has($limits, \tenant::MAILBOX_INSTANCE_SIZE)) {
            return;
        }

        $mailboxCount = Arr::get($limits, \tenant::MAILBOX_COUNT, 0);
        if ($mailboxCount === 0) {
            return;
        }

        if ($this->tenant->type !== \tenant::TYPE_CUSTOMER) {
            if ($this->checkOfferingItemIsEnabledForRoot(Api::OFFERING_ITEM_MAILBOX_INSTANCE_SIZE)) {
                $limits[\tenant::MAILBOX_INSTANCE_SIZE] = Counter::UNLIMITED_VALUE;
            };

            return;
        }

        $tenant = $this->tenant;
        if ($tenant->profile->isEmpty() || !$tenant->profile->freemiumMailboxStorage) {
            return;
        }

        $storageSize = Arr::get($limits, \tenant::STORAGE_USAGE, 0);
        $mailboxInstanceSize = $tenant->profile->freemiumMailboxStorage;

        $limits[\tenant::MAILBOX_INSTANCE_SIZE] = $this->computeMailboxInstanceStorageSizeForTenant(
            $storageSize, $mailboxCount, $mailboxInstanceSize
        );
    }

    /**
     * @param array $limits
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function addO365MailboxesCounter(array &$limits)
    {
        // do nothing if the counter is already specified in the service plan
        if (Arr::has($limits, \tenant::O365_MAILBOXES)) {
            return;
        }

        // do nothing if there are no Microsoft 365 seats
        if (Arr::get($limits, \tenant::MAILBOX_COUNT, 0) === 0) {
            return;
        }

        // do nothing if the offering item is not enabled for root
        if (!$this->checkOfferingItemIsEnabledForRoot(Api::OFFERING_ITEM_O365_MAILBOXES)) {
            return;
        };

        $limits[\tenant::O365_MAILBOXES] = Counter::UNLIMITED_VALUE;
    }

    /**
     * @param array $limits
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function addO365OneDriveCounter(array &$limits)
    {
        // do nothing if the counter is already specified in the service plan
        if (Arr::has($limits, \tenant::O365_ONE_DRIVE)) {
            return;
        }

        // do nothing if there are no Microsoft 365 seats
        if (Arr::get($limits, \tenant::MAILBOX_COUNT, 0) === 0) {
            return;
        }

        // do nothing if the offering item is not enabled for root
        if (!$this->checkOfferingItemIsEnabledForRoot(Api::OFFERING_ITEM_O365_ONEDRIVE)) {
            return;
        };

        $limits[\tenant::O365_ONE_DRIVE] = Counter::UNLIMITED_VALUE;
    }

    /**
     * @param array $limits
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function addO365SharePointSitesCounter(array &$limits)
    {
        // do nothing if the counter is already specified in the service plan
        if (Arr::has($limits, \tenant::O365_SHARE_POINT_SITES_COUNT)) {
            return;
        }

        // do nothing if there are no Microsoft 365 seats
        if (Arr::get($limits, \tenant::MAILBOX_COUNT, 0) === 0) {
            return;
        }

        // do nothing if the offering item is not enabled for root
        if (!$this->checkOfferingItemIsEnabledForRoot(Api::OFFERING_ITEM_O365_SHAREPOINT_SITES)) {
            return;
        };

        $limits[\tenant::O365_SHARE_POINT_SITES_COUNT] = Counter::UNLIMITED_VALUE;
    }

    /**
     * @param array $limits
     */
    private function addLocalStorageUsageCounter(array &$limits)
    {
        // do nothing if the counter is already specified in the service plan
        if (Arr::has($limits, \tenant::LOCAL_STORAGE_USAGE)) {
            return;
        }

        $profile = $this->tenant->profile;
        if ($profile->isEmpty() || !$profile->storageLocation) {
            return;
        }

        $limits[\tenant::LOCAL_STORAGE_USAGE] = Counter::UNLIMITED_VALUE;
    }

    /**
     * @param int|null $storageSize
     * @param int|null $mailboxCount
     * @param int|null $mailboxInstanceSize
     * @return int|null
     */
    private function computeMailboxInstanceStorageSizeForTenant($storageSize, $mailboxCount, $mailboxInstanceSize)
    {
        if ($mailboxCount === Counter::UNLIMITED_VALUE ||
            $storageSize === Counter::UNLIMITED_VALUE ||
            $mailboxInstanceSize === Counter::UNLIMITED_VALUE
        ) {
            return Counter::UNLIMITED_VALUE;
        }

        if ($storageSize > $mailboxCount * $mailboxInstanceSize) {
            return Counter::UNLIMITED_VALUE;
        }

        return $mailboxInstanceSize;
    }
}