<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\CloudApi;

use Acronis\App\Utils\CacheableTrait;
use Acronis\Cloud\Client\ApiException;
use Acronis\Cloud\Client\HttpException;
use Acronis\Cloud\Client\HttpResponseInterface;
use Acronis\Cloud\Client\Model\Idp\Token;

/**
 * Class AuthorizedApi
 * @package Acronis\App\CloudApi
 */
abstract class AuthorizedApi extends AwareApi
{
    use CacheableTrait;

    /**
     * @return string
     */
    abstract public function getLogin();

    /**
     * @return string
     */
    abstract public function getPassword();

    /**
     * @param callable $fn
     * @return mixed
     */
    protected function call(callable $fn)
    {
        /** @var HttpResponseInterface $response */
        $response = $fn();

        if (is_object($response) && $response instanceof HttpResponseInterface) {
            return $response->getResponseBody();
        }

        return $response;
    }

    /**
     * @param callable $fn
     * @return mixed
     */
    protected function unauthorizedCall(callable $fn)
    {
        $clientConfig = $this->getApiClient()->getConfig();

        $accessToken = $clientConfig->getAccessToken();
        $clientConfig->setAccessToken('');

        $response = $this->call($fn);

        $clientConfig->setAccessToken($accessToken);

        return $response;
    }

    /**
     * @param callable $fn
     * @return mixed
     * @throws HttpException
     * @throws ApiException
     */
    protected function authorizedCall(callable $fn)
    {
        if (!$this->isAuthorized()) {
            $this->authorize();
        }
        try {
            return $this->call($fn);
        } catch (HttpException $e) {
            if ($e->getCode() != 401) {
                throw $e;
            }
            $this->authorize(true);
        }

        return $this->call($fn);
    }

    /**
     * @param bool $force
     * @return string
     */
    protected function obtainAccessToken($force = false)
    {
        return $this->fromCache(function () {
            /** @var Token $idpToken */
            $idpToken = $this->unauthorizedCall(function () {
                return $this->getIdpApi()->postIdpToken(
                    'password', null, $this->getLogin(), $this->getPassword()
                );
            });

            return $idpToken->getAccessToken();
        }, $this->getAccessTokenTtl(), true, $this->getCredentialsHash(), $force);
    }

    abstract protected function getCredentialsHash();

    private function isAuthorized()
    {
        $apiToken = $this->getApiClient()
            ->getConfig()
            ->getAccessToken();
        $notificationApiToken = $this->getApiClient()
            ->getConfig()
            ->getAccessToken();

        return !empty($apiToken) and !empty($notificationApiToken);
    }

    private function authorize($force = false)
    {
        $accessToken = $this->obtainAccessToken($force);
        $this->getApiClient()
            ->getConfig()
            ->setAccessToken($accessToken);
        $this->getNotificationApiClient()
            ->getConfig()
            ->setAccessToken($accessToken);
    }
}