<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Console\Command;

use Acronis\App\Base\ApplicationsManager;
use Acronis\App\CloudApi\Api;
use Acronis\App\Services\Apsc\ApscInstanceFactoryTrait;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Acronis\App\Services\Logger\AppLoggerAwareTrait;
use Acronis\App\Utils\MemoizeTrait;
use Acronis\Cloud\Client\Model\OfferingItems\OfferingItemOutput;
use APSResource;
use Arr;
use RqlBuilder;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Time;
use ZipStream\ZipStream;

/**
 * Performs only up sql
 * @package Acronis\App\Console\Command
 */
final class Collect extends Loggable
{
    use AppLoggerAwareTrait,
        ApscInstanceFactoryTrait,
        MemoizeTrait,
        CountersManagerAwareTrait,
        \Service_MspApiClient_MspApiFactoryTrait;

    const ARGUMENT_SUBSCRIPTION_ID = 'subscriptionId';
    const FILE_NAME_PREFIX_TENANT = 'subscription_%s';
    const FILE_NAME_PREFIX_USER = 'subscription_%s_user_%s';
    const ENTRY_NAME = 'subscription_%s_%s';

    private $outputFolderPath = DIR_TMP . 'collected';

    private $files = [];

    protected function configure()
    {
        parent::configure();

        $this->addArgument(
            static::ARGUMENT_SUBSCRIPTION_ID,
            InputArgument::REQUIRED,
            'ID of subscription to collect all related APS resources and Cloud state (optional)'
        );
    }

    protected function getCommandName()
    {
        return 'collect:collect';
    }

    protected function getCommandHelp()
    {
        return '<info>Outputs an archive with Cloud / APS offering items state.</info>';
    }

    protected function getCommandDescription()
    {
        return 'Collects current state of APS / Cloud.';
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        try {
            $subscriptionId = $input->getArgument(static::ARGUMENT_SUBSCRIPTION_ID);
            $tenant = $this->fetchTenant($subscriptionId);
            $api = new Api($tenant->datacenter->url, $tenant->datacenter->login, $tenant->datacenter->password);
            $applicationsIds = $this->getAvailableApplicationsIds($tenant->datacenter);

            $tenantOfferingItems = ['aps' => [], 'cloud' => []];
            $usersOfferingItems = [];

            foreach ($applicationsIds as $applicationId) {
                $tenantOfferingItems['cloud'] = array_merge(
                    $tenantOfferingItems['cloud'],
                    $api->getTenantApplicationOfferingItems($tenant->tenantId, $applicationId)
                );
                /** @var \users $user */
                foreach ($tenant->users as $user) {
                    if (!isset($usersOfferingItems[$user->aps->id])) {
                        $usersOfferingItems[$user->aps->id] = [
                            'cloud' => [],
                            'aps' => $user->offeringItems,
                        ];
                    }
                    $usersOfferingItems[$user->aps->id]['cloud'] = array_merge(
                        $usersOfferingItems[$user->aps->id]['cloud'],
                        $api->getTenantApplicationOfferingItems($user->personalTenantId, $applicationId)
                    );
                }
            }

            foreach ($tenant->subscription->resources() as $resource) {
                if (!isset($resource->property)) {
                    continue;
                }
                $counter = $this->getCountersManager()->getCounterByName($resource->property);
                $tenantOfferingItems['aps'][] = [
                    'name' => $counter->getOfferingItemName(),
                    'limit' => Arr::getNestedItem($tenant, [$counter->getName(), 'limit'], 0),
                    'usage' => Arr::getNestedItem($tenant, [$counter->getName(), 'usage'], 0),
                ];
            }

            $this->saveOfferingItemsState(sprintf(static::FILE_NAME_PREFIX_TENANT, $subscriptionId),
                $tenantOfferingItems);
            foreach ($usersOfferingItems as $userId => $userOfferingItems) {
                $this->saveOfferingItemsState(sprintf(static::FILE_NAME_PREFIX_USER, $subscriptionId, $userId),
                    $userOfferingItems);
            }

            echo $this->writeFiles($subscriptionId) . PHP_EOL;
        } catch (\Exception $e) {
            $this->getLogger()->error(sprintf('Exception during collecting data occurred: "%s".', $e->getMessage()));
        }
    }

    /**
     * @param $rql
     * @return \tenant
     * @throws \Exception
     */
    private function fetchTenant($subscriptionId)
    {
        $rql = RqlBuilder::createInstance()
            ->implementing(APSResource::CLOUD_BACKUP_TENANT_TYPE)
            ->eq('subscriptionId', $subscriptionId)
            ->getQuery();

        $resources = $this->getApsc()->getResources($rql);
        if (count($resources) === 1) {
            /** @var \tenant $tenantRaw */
            $tenantRaw = reset($resources);
            $tenant = new \tenant();
            $tenant->fillResource($tenantRaw->aps->id);

            return $tenant;
        } else {
            throw new \Exception(sprintf('Actual count of tenants matching RQL "%s" is %s, expected 1.',
                $rql, count($resources)
            ));
        }
    }

    /**
     * Returns only first instance
     * @return mixed
     */
    private function getApsc()
    {
        return $this->memoize(function () {
            $instances = \APS\ControllerProxy::listInstances();
            return $this->createApscInstance(reset($instances));
        });
    }

    /**
     * @param $datacenter
     * @return array
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     * @throws \Rest\RestException
     * @throws \Service_BaseClient_HTTPException
     */
    private function getAvailableApplicationsIds($datacenter)
    {
        $mspApi = $this->createMspApiInstance($datacenter->url, $datacenter->login, $datacenter->password);
        $applicationManager = new ApplicationsManager(null, $mspApi, 123);
        $applicationIds = [];
        foreach (ApplicationsManager::AVAILABLE_APPLICATION_TYPES as $applicationType) {
            $applicationIds[] = $applicationManager->getApplicationIdByType($applicationType);
        }

        return $applicationIds;
    }

    private function saveOfferingItemsState($fileNamePrefix, $offeringItems)
    {
        $offeringItemsFromCloud = [];
        foreach ($offeringItems['cloud'] as $offeringItem) {
            if ($offeringItem->getStatus() === 0) {
                continue;
            }

            /** @var OfferingItemOutput $offeringItem */
            $oi = [
                'name' => $offeringItem->getName(),
                'status' => $offeringItem->getStatus(),
            ];
            if ($offeringItem->hasQuota()) {
                $offeringItemQuota = $offeringItem->getQuota();
                $quota = [];
                $offeringItemQuota->hasValue() && $quota['value'] = $offeringItemQuota->getValue();
                $offeringItemQuota->hasOverage() && $quota['overage'] = $offeringItemQuota->getOverage();
                $offeringItemQuota->hasVersion() && $quota['version'] = $offeringItemQuota->getVersion();
                $oi['quota'] = $quota;
            }
            if ($offeringItem->hasInfraId()) {
                $oi['infra_id'] = $offeringItem->getInfraId();
            }
            $offeringItemsFromCloud[] = $oi;
        }

        $this->files[$fileNamePrefix . '_from_aps.json'] = json_encode($offeringItems['aps'], JSON_PRETTY_PRINT);
        $this->files[$fileNamePrefix . '_from_cloud.json'] = json_encode($offeringItemsFromCloud, JSON_PRETTY_PRINT);
    }

    private function writeFiles($subscriptionId)
    {
        try {
            $entryName = sprintf(
                static::ENTRY_NAME,
                $subscriptionId, Time::getCurentDate('UTC', 'Y-m-d_H:i:s.') . round(gettimeofday()["usec"] / 1000)
            );
            if (!file_exists($this->outputFolderPath)) {
                mkdir($this->outputFolderPath);
            }

            $folder = $this->outputFolderPath . '/' . $subscriptionId;
            if (!file_exists($folder)) {
                mkdir($folder);
            }

            ob_start();
            $archivePath =  $folder . '/' . $entryName . '.zip';
            $zipFile = new ZipStream($archivePath);
            foreach ($this->files as $name => $contents) {
                $zipFile->addFile($name, $contents);
            }
            $zipFile->finish();
            file_put_contents($archivePath, ob_get_contents());

            return $archivePath;
        } catch (\Exception $e) {
            return null;
        } finally {
            ob_end_clean();
        }
    }
}