<?php
/**
 * @Copyright © 2002-2021 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Console\Command\UsageReport;

use Acronis\App\Base\GroupIdConverterTrait;
use Acronis\App\Console\Command\Loggable;
use Acronis\App\Services\Apsc\ApscInstanceFactoryTrait;
use Acronis\App\UsageReport\TenantsMap\ResellersAccountToTenantMapBuilder;
use Acronis\App\UsageReport\TenantsMap\TenantMapping;
use Acronis\App\UsageReport\TenantsMap\TenantMappingInterface;
use Acronis\App\UsageReport\TenantsMap\TenantsMap;
use Acronis\App\Utils\CacheableTrait;
use Acronis\App\Utils\MemoizeTrait;
use APS\Proto;
use APSResource;
use Json;
use RqlBuilder;
use Service_MspApiClient_MspApiFactoryTrait;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;

class WarmUp extends Loggable
{
    use ApscInstanceFactoryTrait;
    use CacheableTrait;
    use MemoizeTrait;
    use Service_MspApiClient_MspApiFactoryTrait;
    use GroupIdConverterTrait;

    const COMMAND_NAME = 'usage-report:warm-up';
    const TENANTS_BATCH_SIZE = 1000;

    protected function getCommandName()
    {
        return static::COMMAND_NAME;
    }

    protected function getCommandHelp()
    {
        return <<<TEXT
<info>
This command single execution is needed as a pre-condition before usage reporting in order to place some technical information 
into cache and make whole process more effective and less resource usage intensive.
</info>
TEXT;
    }

    protected function getCommandDescription()
    {
        return 'Usage report caching command.';
    }

    /**
     * @param InputInterface $input
     * @param OutputInterface $output
     * @return int|null|void
     * @throws \Exception
     */
    protected function execute(InputInterface $input, OutputInterface $output)
    {
        foreach (\APS\ControllerProxy::listInstances() as $apsInstanceId) {
            $datacenters = $this->fetchDatacentersForApsInstance($apsInstanceId);

            /** @var \datacenter $datacenter */
            foreach ($datacenters as $datacenter) {
                foreach ($this->fetchTenantMappings($datacenter, $apsInstanceId) as $tenantMapping) {
                    $this->partialFillMap(TenantsMap::NAME, $tenantMapping->getTenantId(), $tenantMapping);
                }
            }
        }

        $this->setMapIsWarmedUpStatus(TenantsMap::NAME);
    }

    /**
     * @param $apsInstanceId
     * @return array
     * @throws \Exception
     */
    private function fetchDatacentersForApsInstance($apsInstanceId)
    {

        $rql = RqlBuilder::createInstance()
            ->implementing(APSResource::CLOUD_BACKUP_DATACENTER_TYPE_MAJOR)
            ->getQuery();

        return $this->getApsc($apsInstanceId)->getResources($rql);
    }

    private function getApsc($apsInstanceId)
    {
        return $this->memoize(function () use ($apsInstanceId) {
            return $this->createApscInstance($apsInstanceId);
        }, $apsInstanceId);
    }

    /**
     * @param $datacenter
     * @param $apsInstanceId
     * @return TenantMappingInterface[]
     * @throws \Exception
     */
    private function fetchTenantMappings($datacenter, $apsInstanceId)
    {
        $this->getLog()->notice('Fetching tenants from datacenter "%s" (%s).', $datacenter->name, $datacenter->aps->id);
        $start = microtime(true);
        $apsQueriesTotalTime = 0;
        $apsc = $this->getApsc($apsInstanceId);
        $path = $apsc->getIo()->resourcePath();
        $accountToTenantMapBuilder = new ResellersAccountToTenantMapBuilder($datacenter,
            $this->getApsc($apsInstanceId));
        $accountToTenantMap = $accountToTenantMapBuilder->build();

        $fetchTenantsResources = function ($offset, $limit) use ($datacenter, $path, &$apsQueriesTotalTime, $apsc) {
            $rqlFilter = RqlBuilder::createInstance()
                ->implementing(\APSResource::CLOUD_BACKUP_TENANT_TYPE)
                ->eq('aps.status', 'aps:ready')
                ->eq('datacenter.aps.id', $datacenter->aps->id)
                ->eq('type', \tenant::TYPE_RESELLER, RqlBuilder::JOIN_OR)
                ->limit($offset, $limit)
                ->select(['tenantId', 'subscriptionId', 'type', 'accountId', 'datacenter'])
                ->getQuery();

            $start = microtime(true);
            $response = $apsc->getIo()->sendRequest(Proto::GET, $path . '?' . $rqlFilter);
            $elapsed = microtime(true) - $start;
            $apsQueriesTotalTime += $elapsed;

            return Json::decode($response);
        };

        $offset = 0;
        while (true) {
            $tenants = $fetchTenantsResources($offset, static::TENANTS_BATCH_SIZE);
            $offset += count($tenants);

            if (!count($tenants)) {
                break;
            }

            /** @var \tenant $tenant */
            foreach ($tenants as $tenant) {
                if ($tenant->type === \tenant::TYPE_RESELLER) {
                    if (!isset($accountToTenantMap[$tenant->accountId])) {
                        $this->getLog()->warning(
                            'There is no reseller relation for account "%s".',
                            $tenant->accountId
                        );
                        continue;
                    }

                    $tenant->tenantId = $accountToTenantMap[$tenant->accountId];
                }

                yield new TenantMapping($tenant->tenantId, (object) [
                    'subscriptionId' => $tenant->subscriptionId,
                    'accountId' => $tenant->accountId,
                    'type' => $tenant->type,
                ]);
            }
        }

        $elapsed = microtime(true) - $start;

        $this->getLog()->notice(
            'Finished fetching tenants mapping for datacenter "%s" (%s). Took %ss, including APSc: %ss.',
            $datacenter->name, $datacenter->aps->id, round($elapsed, 2), round($apsQueriesTotalTime, 2)
        );
    }
}