<?php
/**
 * @Copyright © 2002-2021 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Services\Redis;

use Acronis\App\Services\Locator;
use Acronis\App\Services\Logger\AppLoggerAdapter;
use Acronis\App\Services\Logger\AppLoggerFactory;
use Acronis\App\Services\Redis\Exception\ClientException;
use Predis\Client;
use Str;

class RedisClient
{
    const EXPIRE_RESOLUTION_SECONDS = 'EX';
    const TTL_NO_EXPIRATION = 0;

    /** @var Client */
    private $client;

    /** @var string */
    private $keyPrefix;

    /**
     * RedisClient constructor.
     * @param Client $client
     * @param string $keyPrefix
     */
    public function __construct($client, $keyPrefix = '')
    {
        $this->client = $client;
        $this->keyPrefix = $keyPrefix;
    }

    /**
     * @param string $key
     * @return mixed
     * @throws ClientException
     */
    public function get($key)
    {
        return $this->redisCall(function () use ($key) {
            return $this->client->get($this->formatKey($key));
        });
    }

    /**
     * @param string $key
     * @return boolean
     * @throws ClientException
     */
    public function has($key)
    {
        return $this->redisCall(function () use ($key) {
            return $this->client->exists($this->formatKey($key));
        });
    }

    /**
     * @param string $key
     * @param int|null $ttl
     * @throws ClientException
     */
    public function refresh($key, $ttl)
    {
        $this->redisCall(function () use ($key, $ttl) {
            $this->client->expire($key, $ttl);
        });
    }

    /**
     * @param string $customPrefix
     * @throws ClientException
     */
    public function deleteAll($customPrefix = '')
    {
        $this->redisCall(function () use ($customPrefix) {
            $keys = $this->client->keys($this->getWildCard($customPrefix));
            if (count($keys)) {
                $this->client->del($keys);
            }
        });
    }

    /**
     * @param $key
     * @throws ClientException
     */
    public function delete($key)
    {
        $this->redisCall(function () use ($key) {
            $this->client->del($key);
        });
    }

    /**
     * @param string $customPrefix
     * @return string
     */
    private function getWildCard($customPrefix)
    {
        return $this->formatKey($customPrefix . '*');
    }

    /**
     * @param string $key
     * @param mixed $value
     * @param int|null $ttl
     * @throws ClientException
     */
    public function set(
        $key,
        $value,
        $ttl = self::TTL_NO_EXPIRATION
    ) {
        $this->redisCall(function () use ($key, $value, $ttl) {
            if ($ttl > 0) {
                $this->client->set($this->formatKey($key), $value, static::EXPIRE_RESOLUTION_SECONDS, $ttl);
            } else {
                $this->client->set($this->formatKey($key), $value);
            }
        });
    }

    /**
     * @param string $key
     * @return string
     */
    private function formatKey($key)
    {
        return $this->keyPrefix . $key;
    }

    /**
     * @param callable $redisClosure
     * @throws ClientException
     * @return mixed
     */
    private function redisCall(callable $redisClosure)
    {
        try {
            return $redisClosure();
        } catch (\Exception $e) {
            $errorMessage = Str::format(
                'Unable to reach Redis client. Error: "%s".',
                $e->getMessage()
            );

            /** @var AppLoggerAdapter $log */
            $log = Locator::getInstance()->get(AppLoggerFactory::NAME);
            $log->error($errorMessage);

            throw new ClientException($errorMessage, $e->getCode(), $e);
        }
    }
}