<?php
/**
 * @Copyright © 2002-2021 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\UsageReport\TenantsMap;

use Acronis\App\Base\GroupIdConverterTrait;
use Acronis\App\Services\Logger\AppLoggerAwareTrait;
use Acronis\App\UsageReport\Exception\ManagerException;
use APS\ControllerProxy;
use RqlBuilder;
use RuntimeException;
use Service_BaseClient_HTTPException;

class ResellersAccountToTenantMapBuilder
{
    use AppLoggerAwareTrait;
    use GroupIdConverterTrait;
    use \Service_MspApiClient_MspApiFactoryTrait;

    /**
     * ResellersAccountToTenantMapBuilder constructor.
     * @param \datacenter $datacenter
     * @param ControllerProxy $apsc
     */
    public function __construct(private $datacenter, private $apsc)
    {
    }

    public function build(): array
    {
        $start = microtime(true);
        $this->getLog()->notice(
            'Building resellers account to tenant mapping for DC "%s".',
            $this->datacenter->aps->id
        );

        $accountToTenantMap = [];
        $global = $this->getGlobal();
        $api = $this->createMspApiInstance(
            $this->datacenter->url,
            $this->datacenter->login,
            $this->datacenter->password,
            $this->apsc->getInstanceId()
        );

        /** @var \ResellerRelation[] $resellerRelation */
        $resellerRelation = $global->resellerRelations ?: [];

        foreach ($resellerRelation as $relation) {
            if ($relation->datacenterId !== $this->datacenter->id) {
                continue;
            }

            try {
                $accountToTenantMap[$relation->accountId] = $this->convertToTenantId($relation->groupId, $api);
            } catch (Service_BaseClient_HTTPException $e) {
                if ($e->getCode() != 404) {
                    throw $e;
                }

                $this->getLog()->warning(
                    'Unable to resolve tenant UUID, group ID "%s".',
                    $relation->groupId
                );
            }
        }

        $elapsed = microtime(true) - $start;
        $this->getLog()->notice('Successfully built account to tenant mapping. Took: %s.', round($elapsed, 2));

        return $accountToTenantMap;
    }

    /**
     * @return \globals|null
     * @throws ManagerException
     */
    private function getGlobal()
    {
        $rqlFilter = RqlBuilder::createInstance()
            ->implementing(\APSResource::CLOUD_BACKUP_GLOBALS_TYPE)
            ->getQuery();
        /** @var \globals[]|null $globals */
        $globals = $this->apsc->getResources($rqlFilter);

        if (empty($globals)) {
            throw new RuntimeException('Unable to fetch reseller relation information.');
        }

        return reset($globals);
    }
}