<?php
/**
 * @Copyright © 2002-2021 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\UsageReport\TenantsMap;

use Acronis\App\Base\GroupIdConverterTrait;
use Acronis\App\Base\SimpleCompositeKey;
use Acronis\App\Services\Apsc\ApscInstanceFactoryTrait;
use Acronis\App\Services\Config\ConfigAwareTrait;
use Acronis\App\Services\Config\ConfigSchema;
use Acronis\App\Services\Logger\AppLoggerAwareTrait;
use Acronis\App\UsageReport\Exception\ManagerException;
use Acronis\App\UsageReport\Exception\TenantsMapException;
use Acronis\App\Utils\CacheableTrait;
use Acronis\App\Utils\MemoizeTrait;
use APS\ControllerProxy;
use APS\Proto;
use Json;
use RqlBuilder;
use Service_BaseClient_HTTPException;
use Service_MspApiClient_MspApiFactoryTrait;

class TenantsMap implements TenantsMapInterface
{
    use AppLoggerAwareTrait;
    use CacheableTrait;
    use GroupIdConverterTrait;
    use Service_MspApiClient_MspApiFactoryTrait;
    use MemoizeTrait;
    use ApscInstanceFactoryTrait;
    use ConfigAwareTrait;

    const NAME = 'tenants_map';

    /**
     * TenantsMap constructor.
     * @param \globals $global
     * @param ControllerProxy $apsc
     * @param \datacenter $datacenter
     * @param int $reportEntryId
     */
    public function __construct(private $global, private $apsc, private $datacenter, private $reportEntryId)
    {
    }

    public function offsetExists($offset)
    {
        return $this->getTenantMappingByTenantId($offset) !== null;
    }

    public function getTenantMappingByTenantId(string $tenantId) : ?TenantMappingInterface
    {
        return $this->fromCacheMap(static::NAME, $tenantId, function () use ($tenantId) {
            return $this->fetchTenantById($tenantId);
        });
    }

    /**
     * @param $tenantId
     * @return ?TenantMappingInterface
     * @throws ManagerException
     * @throws Service_BaseClient_HTTPException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function fetchTenantById($tenantId)
    {
        // todo make it work for reseller
        $tenantToResellerAccountMap = $this->buildTenantToResellerAccountMap();

        if (isset($tenantToResellerAccountMap[$tenantId])) {
            $rqlFilter = RqlBuilder::createInstance()
                ->implementing(\APSResource::CLOUD_BACKUP_TENANT_TYPE)
                ->eq('aps.status', 'aps:ready')
                ->eq('accountId', $tenantToResellerAccountMap[$tenantId])
                ->select(['tenantId', 'subscriptionId', 'type', 'accountId', 'datacenter'])
                ->getQuery();
            $tenant = $this->fetchFirstTenant($rqlFilter);

            return new TenantMapping($tenantId, (object) [
                'subscriptionId' => $tenant->subscriptionId,
                'accountId' => $tenant->accountId,
                'type' => $tenant->type,
            ]);
        }

        $rqlFilter = RqlBuilder::createInstance()
            ->implementing(\APSResource::CLOUD_BACKUP_TENANT_TYPE)
            ->eq('aps.status', 'aps:ready')
            ->eq('tenantId', $tenantId)
            ->select(['tenantId', 'subscriptionId', 'type', 'accountId', 'datacenter'])
            ->getQuery();

        try {
            $tenant = $this->fetchFirstTenant($rqlFilter);

            return new TenantMapping($tenantId, (object) [
                'subscriptionId' => $tenant->subscriptionId,
                'accountId' => $tenant->accountId,
                'type' => $tenant->type,
            ]);
        } catch (\Exception) {
            return null;
        }
    }

    /**
     * @param \datacenter $datacenter
     * @param $apsInstanceId
     * @return object[]
     * @throws ManagerException
     * @throws \Service_BaseClient_HTTPException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function buildTenantToResellerAccountMap()
    {
        return $this->memoize(function () {
            $accountToTenantMapBuilder = new ResellersAccountToTenantMapBuilder($this->datacenter, $this->apsc);

            return array_flip($accountToTenantMapBuilder->build());
        }, new SimpleCompositeKey([$this->datacenter->aps->id, $this->apsc->getInstanceId()]));
    }

    /**
     * @param mixed $rqlFilter
     * @return mixed
     * @throws \Exception
     */
    private function fetchFirstTenant(mixed $rqlFilter)
    {
        $path = $this->getApsc()->getIo()->resourcePath();
        $response = $this->getApsc()->getIo()->sendRequest(Proto::GET, $path . '?' . $rqlFilter);
        $tenants = Json::decode($response);

        if (is_array($tenants)) {
            return array_first($tenants);
        }

        return null;
    }

    private function getApsc() : ControllerProxy
    {
        return $this->apsc;
    }

    public function offsetGet($offset)
    {
        return $this->getTenantMappingByTenantId($offset);
    }

    public function offsetSet($offset, $value)
    {
        throw new TenantsMapException('Not supposed to be called.');
    }

    public function offsetUnset($offset)
    {
        throw new TenantsMapException('Not supposed to be called.');
    }

    protected function warmIsRequired() : bool
    {
        return $this->getConfig()->getUsageReportConfig()[ConfigSchema::NODE_USAGE_REPORT_REQUIRE_CACHE_WARM_UP] ?? false;
    }
}