<?php
/**
 * @Copyright © 2002-2021 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Utils;

use Acronis\App\Services\Config\ConfigSchema;
use Acronis\App\Services\Cache\{CacheFactory, CacheInterface};
use Acronis\App\Services\Locator;
use Acronis\App\Services\Logger\{AppLoggerAdapter, AppLoggerFactory};
use Acronis\App\Utils\Exception\CachedMapIsNotWarmedUp;

/**
 * Trait CacheableTrait
 * Use CacheableTrait in cases you want to fetch/build some object identified by $objectKey only one time and then
 * return it from cache every time when possible
 * $this->fromCache(function() use ($anyData) { return $this->veryExpensiveFetchRequest($objectKey); }, objectKey, [force]);
 * If you want to be able to fallback to fetching/building object every time, pass last parameter as true
 * @package Acronis\App\Utils
 */
trait CacheableTrait
{
    protected string $mapPrefix = 'cached_map';

    /**
     * Use this method in order to cache some result, by id, also it will
     * add uniqueness prefix based on context of execution (class FQDN + caller name)
     */
    protected function fromCache(
        callable $method,
        ?int $ttl = null,
        bool $refreshTtl = false,
        \Stringable|string $id = '',
        bool $force = false
    ) {
        $callerName = Func::getCallerName();
        $objectId = $this->buildCacheableObjectKey($callerName, $id);

        return $this->getObjectFromCache($objectId, $method, $ttl, $force, $refreshTtl);
    }

    /**
     * Use this method in order to cache mappings in pair with setMapIsFilled($mapName).
     * The main idea is that effective map should be filled in background and only after that
     * it can be used, of course all new mappings may be added on fly.
     * "map" is just a scope of key-values in cache having same key prefixes, nothing more.
     * @throws CachedMapIsNotWarmedUp
     * @see setMapIsWarmedUpStatus
     */
    protected function fromCacheMap(string $mapName, \Stringable|string $id, callable $method, $ttl = null)
    {
        if ($this->warmIsRequired() && !$this->fetchMapIsWarmedUp($mapName)) {
            throw new CachedMapIsNotWarmedUp($mapName, $id);
        }

        $objectId = $this->buildCacheableObjectKeyForMap($mapName, $id);

        return $this->getObjectFromCache($objectId, $method, $ttl, false, false);
    }

    protected function partialFillMap(string $mapName, \Stringable|string $id, $value, $ttl = null)
    {
        $objectId = $this->buildCacheableObjectKeyForMap($mapName, $id);
        $this->storeCacheableObject($objectId, $value, $ttl);
    }

    protected function setMapIsWarmedUpStatus($mapName)
    {
        $this->getCacheInstance()->set($this->buildMapIsWarmedUpKey($mapName), true);
    }

    protected function clearMap($mapName)
    {
        $this->getCacheInstance()->deleteAll($this->buildCacheableObjectKeyForMap($mapName, ''));
    }

    protected function clearMapIsWarmedUpStatus($mapName)
    {
        $this->getCacheInstance()->delete($this->buildMapIsWarmedUpKey($mapName));
    }

    /**
     * @param string $objectId
     * @param callable $method
     * @param int|null $ttl
     * @param bool $force
     * @param bool $refreshTtl
     * @return mixed
     */
    protected function getObjectFromCache(string $objectId, callable $method, ?int $ttl, bool $force, bool $refreshTtl)
    {
        try {
            if (!$force && $this->hasCacheableStoredObject($objectId)) {
                return $this->getCacheableStoredObject($objectId, $ttl, $refreshTtl);
            }
        } catch (\Exception $e) {
            /** @var AppLoggerAdapter $log */
            $log = Locator::getInstance()->get(AppLoggerFactory::NAME);
            $log->error(
                'Unable to fetch object with id "%s" from cache. Error: %s',
                $objectId, $e->getMessage()
            );

            return $method();
        }

        $object = $method();
        $this->storeCacheableObject($objectId, $object, $ttl);

        return $object;
    }

    protected function warmIsRequired(): bool
    {
        // to implement based on cache map logic
        return false;
    }

    /**
     * @param string $objectId
     * @return bool
     */
    private function hasCacheableStoredObject($objectId)
    {
        return $this->getCacheInstance()->has($objectId);
    }

    /**
     * @return CacheInterface
     */
    private function getCacheInstance()
    {
        return Locator::getInstance()->get(CacheFactory::NAME);
    }

    /**
     * @param string $objectId
     * @param int|null
     * @param bool $refreshTtl
     * @return mixed
     */
    private function getCacheableStoredObject($objectId, $ttl, $refreshTtl)
    {
        $result = $this->getCacheInstance()->get($objectId);

        if ($refreshTtl && $this->getCacheInstance()->has($objectId)) {
            $this->getCacheInstance()->refresh($objectId, $ttl);
        }

        return $result;
    }

    /**
     * @param int|null $ttl
     * @param string $objectId
     * @param mixed $value
     */
    private function storeCacheableObject($objectId, $value, $ttl)
    {
        $this->getCacheInstance()->set($objectId, $value, $ttl);
    }

    private function fetchMapIsWarmedUp($mapName) : bool
    {
        return $this->getCacheInstance()->get($this->buildMapIsWarmedUpKey($mapName)) ?? false;
    }

    private function buildMapIsWarmedUpKey(string $mapName) : string
    {
        return "{$this->mapPrefix}_{$mapName}|is_warmed_up";
    }

    private function buildCacheableObjectKeyForMap(string $mapName, \Stringable|string $id) : string
    {
        return "{$this->mapPrefix}_{$mapName}|{$id}";
    }

    private function buildCacheableObjectKey(string $method, \Stringable|string $id) : string
    {
        return "$method|$id";
    }
}