<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

use Acronis\App\Base\ApplicationsManager;
use Acronis\App\Base\SubscriptableInterface;
use Acronis\App\CloudApi\Api;
use Acronis\App\Services\NotificationManager\NotificationManager;
use Acronis\App\Services\NotificationManager\NotificationManagerAwareTrait;
use Acronis\App\Services\AgentsManager\AgentsManager;
use Acronis\App\Services\CountersManager\CountersManager;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Acronis\App\Utils\CacheableTrait;
use Acronis\App\Utils\MemoizeTrait;
use Acronis\Cloud\Client\Model\Common\Types\Role;
use Acronis\Cloud\Client\Model\OfferingItems\OfferingItemOutput;
use Acronis\Cloud\Client\Model\Users\UserPut;
use Model_PropertyConverters_UsersConverterHelper as UsersConverterHelper;
use Rest\RestException;
use Acronis\Cloud\Client\Model\AccessPolicies\AccessPolicy;
use Acronis\Cloud\Client\Model\AccessPolicies\AccessPolicyTrusteeType;
use Acronis\Cloud\Client\Model\Common\Contact\Contact;
use Acronis\Cloud\Client\Model\Users\UserPost;
use Service_OpenApi_ApiInterface as OssApiInterface;

require_once(__DIR__ . '/app/bootstrap.php');

/**
 * Class users
 * @type("http://acronis.com/aps/cloud-backup-v1/users/7.13")
 * @implements("http://aps-standard.org/types/core/resource/1.0")
 * @implements("http://aps-standard.org/types/core/user/service/1.0")
 */
class users extends APSResource implements SubscriptableInterface
{
    use CacheableTrait,
        CountersManagerAwareTrait,
        MemoizeTrait,
        Model_Resource_DefaultUserInitializationTrait,
        NotificationManagerAwareTrait;

    const TENANT_APPLICATIONS_CACHE_TTL = 60;

    const LOG_FORMAT_PROVISION_START = 'Provision service from customer subscription (%s) for user (%s).';
    const LOG_MIGRATION_WAS_STARTED = 'User migration was started.';
    const LOG_FORMAT_ATTEMPT_TO_CREATE_EXISTING_MSP_USER = 'Attempt to create user (%s) for subscription (%s), which exists in msp';
    const LOG_FORMAT_CREATING_USER_TASK = 'Creating user (%s) for customer subscribtion (%s)';
    const LOG_FORMAT_DELETE_USER_FAILED = 'Delete user (%) failed with message: %s';

    const EXCEPTION_USER_ALREADY_EXISTS = 'The account with email address already exists in the Cyber Protection service. Delete the old account or specify a different service user.';
    const EXCEPTION_USER_ALREADY_ACTIVATED = 'User already activated';
    const EXCEPTION_PASSWORD_MUST_BE_PRESENT = 'Password must be present';
    const EXCEPTION_ERROR_WHILE_USER_DELETING = 'Error while deleting user (%s). Please try again.';
    const EXCEPTION_CHANGE_STATUS_NOT_ACTIVATED = 'Can\'t change account status on not activated user.';
    const EXCEPTION_TENANT_HAS_NOT_LINKED_WITH_DATACENTER_MESSAGE = 'Tenant has not linked with a datacenter.';

    const ACCOUNT_ENABLED = '1';
    const ACCOUNT_DISABLED = '0';

    const BACKUP_LIST_NAME = 'name';
    const BACKUP_LIST_STATUS = 'status';
    const BACKUP_LIST_LAST_BACKUP = 'last_backup';
    const BACKUP_LIST_NEXT_BACKUP = 'next_backup';
    const BACKUP_LIST_TYPE = 'type';
    const STATISTICS_USER_STORAGE_USAGE = 'userStorageUsage';
    const STATISTICS_USER_SERVER_COUNT = 'userServerCount';
    const STATISTICS_USER_VM_COUNT = 'UserVmCount';
    const STATISTICS_USER_WORKSTATION_COUNT = 'UserWorkstationCount';
    const STATISTICS_USER_MOBILE_DEVICE_COUNT = 'UserMobileDeviceCount';
    const STATISTICS_USER_WEB_HOSTING_SERVER_COUNT = 'userWebHostingServerCount';
    const STATISTICS_USER_WEBSITE_COUNT = 'UserWebsiteCount';
    const STATISTICS_USAGE = 'usage';
    const STATISTICS_LIMIT = 'limit';
    const USER_STATUS_NOT_ACTIVATED = 3;
    const USER_STATUS_IN_PROGRESS = 2;
    const USER_STATUS_ACTIVE = 1;
    const USER_STATUS_INACTIVE = 0;
    const USER_STATUS_ERROR = -1;
    const CLIENTS = 'clients';
    const STATISTICS = 'statistics';

    const LOG_CALLED_CONFIGURE = 'Configure user (%s) in subscription (%s) was called.';
    const LOG_CALLED_PROVISION = 'User provision was called.';
    const LOG_CALLED_PROVISION_ASYNC = 'Asynchronous provision for user (%s) from subscription (%s) was called.';
    const LOG_CALLED_UNPROVISION = 'Unprovision service from customer subscription (%s) for user (%s) was called.';
    const LOG_CALLED_RETRIEVE = 'User retrieve was called.';
    const LOG_CALLED_GET_CLIENT = 'Get client was called.';
    const LOG_CALLED_GET_MACHINE_USAGE = 'Get machine usage was called.';
    const LOG_CALLED_GET_WEB_RESTORE_DIGEST = 'Get Web restore digest was called.';
    const LOG_CALLED_GET_TENANT = 'Get tenant was called.';
    const LOG_CALLED_GET_USER = 'Get user was called.';
    const LOG_CALLED_UPDATE_STATISTICS = 'Update statistics was called.';
    const LOG_CALLED_ON_USER_CHANGE = 'Update user (%s) on HS user change in subscription (%s) was called.';
    const LOG_CALLED_ACTIVATE_ACCOUNT = 'Activate account was called.';
    const LOG_CALLED_CHANGE_PASSWORD = 'Change password was called.';
    const LOG_CALLED_GET_PARTNER_CREDENTIALS = 'Get partner credentials was called.';
    const LOG_CALLED_UPGRADE = 'User (%s) upgrade was called';
    const LOG_CALLED_SYNC = 'User (%s) sync was called';
    const LOG_CALLED_GET_AGENTS = 'Get agents was called.';
    const LOG_METHOD_CALL = 'Resource: %s Account: %s Subscription: %s Call: %s';

    const ADMINISTRATOR_PRIVILEGES_NO = 0;
    const ADMINISTRATOR_PRIVILEGES_YES = 1;
    const NOTIFICATIONS_EMPTY = ['false'];

    const URL_REDIRECT_MASK = '%s?jwt=%s&to=backup';

    const STAFF_MEMBER_AVAILABLE = 'Staff member available %s';

    const CACHE_KEY_FOR_APPLICATION_OI = "cloud_app_id_ois_%s_%s";

    const CACHE_KEY_FOR_LAST_USER_UPDATED_TIME = 'last_user_updated_time_%s';

    const USER_UPDATED_CACHE_TTL = 60;

    const APPLICATION_USER_ROLE_MAP = [
        'backup' => Role::BACKUP_USER,
        'notary' => Role::NOTARY_USER,
        'files_cloud' => Role::SYNC_SHARE_USER,
        'omnivoice' => Role::OMNIVOICE_USER
    ];

    const SUBSCRIPTIABLE_NOTIFICATIONS = [
		'anti_malware' => 'com.acronis.protection.amp.notifications'
	];

    /* ======= RESOURCE PROPERTIES ======= */

    /**
     * @var tenant
     * @link("http://acronis.com/aps/cloud-backup-v1/tenant/9.0")
     * @required
     */
    public $tenant;

    /**
     * @link("http://parallels.com/aps/types/pa/user/1.0")
     * @required
     */
    public $user;

    /**
     * @link("http://aps-standard.org/types/core/admin-user/1.0[]");
     * @deprecated
     */
    public $adminusers;

    /**
     * @type(UserOfferingItem[])
     */
    public $offeringItems;

    /**
     * @type(string[])
     */
    public $notifications;

    /**
     * @type(string[])
     */
    public $serviceRoles;

    /**
     * @readonly
     * @type(string)
     */
    public $userId;

    /**
     * @readonly
     * @type(string)
     */
    public $personalTenantId;

    /**
     * @type(string)
     */
    public $login;

    /**
     * true - sync password from service user account
     * false - not synchronize
     * @type(boolean)
     */
    public $activateDuringCreation = true;

    /**
     * @type(string)
     * @title("Login")
     * @Description("End user login")
     */
    public $endUserLogin;

    /**
     * @type(string)
     * @encrypted
     */
    public $endUserPassword;

    /**
     * @type(string)
     * @readonly
     */
    public $endUserEmail;

    /**
     * @type(string)
     * @readonly
     */
    public $displayName;

    /**
     * @type(string)
     * @readonly
     * @title("User Status")
     * @Description("User Status")
     */
    public $userstatus;

    /**
     * @type(string)
     * @title("Service Status")
     * @Description("Service Status")
     */
    public $servicestatus;

    /**
     * @type("Counter")
     */
    public $userStorageUsage;

    /**
     * @type("Counter")
     * @deprecated
     */
    public $userLocalStorageUsage;

    /**
     * @type("Counter")
     */
    public $userServerCount;

    /**
     * @type("Counter")
     */
    public $UserVmCount;

    /**
     * @type("Counter")
     */
    public $UserWorkstationCount;

    /**
     * @type("Counter")
     */
    public $UserMobileDeviceCount;

    /**
     * @type("Counter")
     * @deprecated
     */
    public $UserMailboxCount;

    /**
     * @type("Counter")
     */
    public $userWebHostingServerCount;

    /**
     * @type("Counter")
     */
    public $UserWebsiteCount;

    /**
     * @type(string)
     * @title("Service User ID")
     **/
    public $serviceUserId;

    /**
     * @type(string)
     * @title("Service subscription ID")
     **/
    public $subscriptionId;

    /**
     * @type(integer)
     * @deprecated
     */
    public $mspUserId;

    /**
     * @type(integer)
     * @deprecated
     */
    public $notifyBackup;

    /**
     * @type(boolean)
     * @deprecated
     */
    public $notificationsManagement;

    /**
     * @type(integer)
     */
    public $privileged;

    /**
     * @type(integer)
     */
    public $roles = null;

    /* ======= OTHER PROPERTIES ======= */

    /** @var Model_UserManagerInterface */
    private $_userManager;

    /** @var Model_UserScreenNotificationManagerInterface */
    private $notificationManager;

    protected $eventHandlers = [
        \APS\EventSubscription::Available => 'onUserAvailable',
        \APS\EventSubscription::Changed => 'onUserChange',
    ];

    /* ======= RESOURCE METHODS ======= */

    public function shouldBeSubscribedToEvent($eventType, $context)
    {
        if ($context === 'users::unprovision') {
            return false;
        }
        if ($eventType === \APS\EventSubscription::Available) {
            return $this->userstatus === static::USER_STATUS_ERROR;
        }

        return true;
    }

    public function getResourceIdForEvent($eventType)
    {
        if ($eventType === \APS\EventSubscription::Changed) {
            return $this->user->aps->id;
        }

        return $this->aps->id;
    }

    public function provision()
    {
        try {
            $this->provisionUser();
        } catch (\Rest\Accepted $exception) {
            // do nothing
        }

        // finally section
        try {
            $notificationManager = $this->getScreenNotificationManager();
            $notificationManager->provisionHasStarted();
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
        }

        if (isset($exception)) {
            throw $exception;
        }
    }

    private function provisionUser()
    {
        Log::getInstance()->event(self::LOG_CALLED_PROVISION);

        if (is_null($this->endUserLogin)) {
            $this->endUserLogin = $this->user->login;
        }
        if ($this->activateDuringCreation && is_null($this->endUserPassword)) {
            $this->endUserPassword = $this->user->password;
        }
        $this->subscriptionId = $this->tenant->subscriptionId;
        $this->serviceUserId = $this->user->aps->id;
        $this->displayName = $this->user->displayName;
        $this->userstatus = self::USER_STATUS_IN_PROGRESS;
        $this->servicestatus = self::ACCOUNT_ENABLED;
        $this->initializeCounters();
        $this->convertProperties($this, 'provision');

        Log::getInstance()->event(sprintf(
            self::LOG_FORMAT_PROVISION_START,
            $this->subscriptionId,
            $this->endUserLogin
        ));

        $this->applyProfileUserQuotaMode($this);

        $this->throwRestAccepted();
    }

    public function provisionAsync()
    {
        if (!$this->tenant->ismigration && $this->isServiceUserHasSubscription()) {
            $this->userstatus = self::USER_STATUS_ERROR;
            $this->refreshEventSubscriptions();
            return;
        }
        try {
            $this->provisionUserAsync();
        } catch (\Rest\Accepted $e) {
            // continue async
            throw $e;
        } catch (\Exception $exception) {
            Log::getInstance()->error($exception->getMessage());
            // do nothing
        }

        // finally section
        try {
            $notificationManager = $this->getScreenNotificationManager();
            if (isset($exception)) {
                $notificationManager->provisionHasFailed();
            } else {
                $notificationManager->provisionHasFinished();
            }
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
        }

        if (isset($exception)) {
            if ($this->tenant->ismigration) {
                throw $exception;
            }

            $this->userstatus = self::USER_STATUS_ERROR;
            $this->refreshEventSubscriptions();
            // Do not throw exception
            // Remove user in onUserAvailable
        }
    }

    private function provisionUserAsync()
    {
        Log::getInstance()->event(sprintf(
            self::LOG_CALLED_PROVISION_ASYNC,
            $this->endUserLogin,
            $this->subscriptionId
        ));

        if (!$this->tenant->ismigration) {
            $this->createCloudUser();
            $this->updateUserRoles();
            $this->updatePersonalTenantQuotas();
        } else {
            $this->updateMspUser();
        }

        $this->userstatus = self::USER_STATUS_NOT_ACTIVATED;
        $this->_activateDuringCreation();
        $this->refreshEventSubscriptions();
    }

    /**
     * @param $time
     * @retrun void
     */
    public function setLastUserUpdatedTime($time)
    {
        $cacheKey = sprintf(self::CACHE_KEY_FOR_LAST_USER_UPDATED_TIME, $this->aps->id);
        $this->getCache()->set($cacheKey, $time);
    }

    /**
     * @return integer|null
     */
    public function getLastUserUpdatedTime()
    {
        $cacheKey = sprintf(self::CACHE_KEY_FOR_LAST_USER_UPDATED_TIME, $this->aps->id);
        return $this->getCache()->get($cacheKey, null);
    }

    /**
     * @param users $new
     * @throws Exception
     */
    public function configure($new = null)
    {
        Log::getInstance()->event(sprintf(
            self::LOG_CALLED_CONFIGURE,
            $this->subscriptionId,
            $this->endUserLogin
        ));

        $this->setLastUserUpdatedTime(time());

        if (is_null($new)) {
            return;
        }

        $managerUser = $this;
        if (!$this->lazyGetUserId()) {
            $this->userId = $new->userId;
            $managerUser = $new;
        }

        $notificationManager = $this->getScreenNotificationManager($managerUser);
        $notificationManager->configureHasStarted();
        try {
            if ($this->servicestatus != $new->servicestatus && $this->userstatus == self::USER_STATUS_NOT_ACTIVATED) {
                throw new \Rest\RestException(409, self::EXCEPTION_CHANGE_STATUS_NOT_ACTIVATED);
            }

            $userInfo = $this->getUserInfo(true);
            $version = $userInfo->getVersion();
            $isEnabled = $userInfo->getEnabled();
            $updateStatus = null;
            if (!$isEnabled) {
                $updateStatus = self::ACCOUNT_ENABLED;
            }
            if ($this->userstatus != self::USER_STATUS_NOT_ACTIVATED) {
                if ($this->user->disabled || $new->servicestatus == self::ACCOUNT_DISABLED) {
                    $updateStatus = $this->userstatus = self::ACCOUNT_DISABLED;
                } else {
                    $this->userstatus = self::ACCOUNT_ENABLED;
                }
            }

            $new->notifications = $this->getNotifications($new);
            $this->applyProfileUserQuotaMode($new);
            $this->convertProperties($new, 'configure');
            $this->getCloudApi()
                ->updateUser(
                    $this->lazyGetUserId(),
                    $this->_getUpdateCloudUserInfo($new, $version, $updateStatus)
                )->getVersion();
			$this->updateUserSubscriptiableNotifications($new);

            $this->serviceRoles = $new->serviceRoles;
            $this->offeringItems = $new->offeringItems;
            $this->updateUserRoles();
            $this->updatePersonalTenantQuotas();

            $this->privileged = $new->privileged;
            $this->servicestatus = $new->servicestatus;
            $this->notifications = $new->notifications;
            $this->notifyBackup = $new->notifyBackup;
            $this->notificationsManagement = $new->notificationsManagement;

            $this->initializeCounters();

            foreach (static::getAllCountersNames() as $counter) {
                if (!isset($new->{$counter}, $new->{$counter}->limit)) {
                    continue;
                }
                $this->{$counter}->limit = $new->{$counter}->limit;
            }
        } catch (\Exception $e) {
            $notificationManager->configureHasFailed();
            throw $e;
        }
        $this->refreshEventSubscriptions();
        $notificationManager->configureHasFinished();
    }

    public function unprovision()
    {
        Log::getInstance()->event(sprintf(
            self::LOG_CALLED_UNPROVISION,
            $this->subscriptionId,
            $this->endUserLogin
        ));

        if ($this->userstatus == self::USER_STATUS_ERROR) {
            $this->unprovisionUser();
            return;
        }

        $notificationManager = $this->getScreenNotificationManager();
        $notificationManager->unprovisionHasStarted();
        try {
            $this->unprovisionUser();
        } catch (Exception $e) {
            $notificationManager->unprovisionHasFailed();
            throw $e;
        }

        $this->refreshEventSubscriptions();
        $notificationManager->unprovisionHasFinished();
    }

    /**
     * @throws Service_BaseClient_HTTPException
     */
    private function unprovisionUser()
    {
        Log::getInstance()->event(sprintf(self::LOG_CALLED_UNPROVISION, $this->subscriptionId, $this->endUserLogin));
        if (is_null($this->lazyGetUserId())) {
            return;
        }

        if ($this->tenant->ismigration) {
            return;
        }

        try {
            $userInfo = $this->getMspApi()->getUserById($this->lazyGetUserId());
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() == 404) {
                return;
            } else {
                throw $e;
            }
        }
        if (is_null($userInfo)) {
            return;
        }
        $version = $userInfo->getVersion();

        if ($userInfo->getStatus() == 1) {
            $userManager = $this->_getUserManager();
            $version = $userManager->disableUser($this->lazyGetUserId(), $version);
        }
        try {
            $this->getMspApi()->deleteUser($this->lazyGetUserId(), $version);
        } catch (Exception $exception) {
            Log::getInstance()->error(sprintf(self::LOG_FORMAT_DELETE_USER_FAILED, $exception->getMessage(),
                $this->endUserLogin));
            throw new Exception(sprintf(self::EXCEPTION_ERROR_WHILE_USER_DELETING, $this->endUserLogin));
        }
    }

    public function retrieve()
    {
        Log::getInstance()->event(self::LOG_CALLED_RETRIEVE);
        $userInfo = $this->getMspApi()->getUserById($this->lazyGetUserId());
        $this->initializeCounters();

        $this->userStorageUsage->usage = $userInfo->getUsage()->getStorageSize();
        $this->userStorageUsage->limit = $userInfo->getPrivileges()->getStorageSize();
        $this->userServerCount->usage = $userInfo->getUsage()->getServerCount();
        $this->userServerCount->limit = $userInfo->getPrivileges()->getServerCount();
        $this->UserVmCount->usage = $userInfo->getUsage()->getVmCount();
        $this->UserVmCount->limit = $userInfo->getPrivileges()->getVmCount();
        $this->UserWorkstationCount->usage = $userInfo->getUsage()->getWorkstationCount();
        $this->UserWorkstationCount->limit = $userInfo->getPrivileges()->getWorkstationCount();
        $this->UserMobileDeviceCount->usage = $userInfo->getUsage()->getMobileCount();
        $this->UserMobileDeviceCount->limit = $userInfo->getPrivileges()->getMobileCount();
        $this->userWebHostingServerCount->usage = $userInfo->getUsage()->getWebHostingServerCount();
        $this->userWebHostingServerCount->limit = $userInfo->getPrivileges()->getWebHostingServerCount();
        $this->UserWebsiteCount->usage = $userInfo->getUsage()->getWebsiteCount();
        $this->UserWebsiteCount->limit = $userInfo->getPrivileges()->getWebsiteCount();
    }

    /* ======= RESOURCE EVENTS ======= */

    /**
     * @verb(POST)
     * @path("/onUserAvailable")
     *
     * @param("http://aps-standard.org/types/core/resource/1.0#Notification",body)
     */
    public function onUserAvailable($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        Log::getInstance()->event(sprintf(self::STAFF_MEMBER_AVAILABLE, $this->endUserLogin));

        if ($this->userstatus == self::USER_STATUS_ERROR) {
            $this->_apsc->unprovisionResource($event->source->id);
        }
        $this->refreshEventSubscriptions([\APS\EventSubscription::Available => false]);
    }

    /**
     * @verb(POST)
     * @path("/onUsrChange")
     *
     * @param("http://aps-standard.org/types/core/resource/1.0#Notification",body)
     */
    public function onUserChange($event)
    {
        Log::getInstance()->event(sprintf(
            self::LOG_CALLED_ON_USER_CHANGE,
            $this->endUserLogin,
            $this->subscriptionId
        ));

        if ($this->userstatus != self::USER_STATUS_ACTIVE && $this->userstatus != self::USER_STATUS_INACTIVE) {
            return;
        }

        $userManager = $this->_getUserManager();
        $mspApi = $this->getMspApi();
        $mspUserInfo = $this->getMspUser();
        $version = $mspUserInfo->getVersion();
        $status = $mspUserInfo->getStatus();
        if ($status == Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_DISABLE) {
            $version = $userManager->enableUser($this->lazyGetUserId(), $version);
        }

        if ($this->activateDuringCreation && !$this->isServiceUserAdministrator()) {
            $endUserPassword = $this->user->password;
            if (isset($endUserPassword) && ($this->endUserPassword !== $endUserPassword)) {
                $this->endUserPassword = $endUserPassword;
                $userManager->resetUserPassword($this->endUserLogin, $this->endUserPassword);
            }
        }

        if ($this->user->email != $mspUserInfo->getEmail()) {
            $mspApi->mailUpdate($this->endUserLogin, $this->user->email);
        }

        $updateUserInfo = new Service_MspApiClient_Request_Struct_UpdateUserInfo();
        $status = !$this->user->disabled && $this->servicestatus != self::ACCOUNT_DISABLED ?
            $updateUserInfo::STATUS_ENABLE :
            $updateUserInfo::STATUS_DISABLE;
        $updateUserInfo->setStatus($status);
        $updateUserInfo->setFirstName($this->user->givenName ? $this->user->givenName : " ");
        $updateUserInfo->setLastName($this->user->familyName ? $this->user->familyName : " ");

        $this->setupUserInfoLocale($updateUserInfo);

        $mspApi->updateUser($this->lazyGetUserId(), $updateUserInfo, $version);

        $this->userstatus = $status;
        $this->displayName = $this->user->displayName;
        $this->update();
    }

    /* ======= RESOURCE OPERATIONS ======= */

    /**
     * @verb(GET)
     * @path("/getTenant")
     */
    public function getTenant()
    {
        Log::getInstance()->event(self::LOG_CALLED_GET_TENANT);
        return $this->tenant->toArray();
    }

    /**
     * @verb(GET)
     * @path("/getclient")
     * @param(object)
     */
    public function getclient($language)
    {
        Log::getInstance()->event(self::LOG_CALLED_GET_CLIENT);

        return array(
            // load agent links directly from BaaS backup console
            self::CLIENTS => $this->getAgents($language),
            self::STATISTICS => $this->updateStatistics()
        );
    }

    /**
     * @verb(GET)
     * @path("/getUser")
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    public function getUser()
    {
        Log::getInstance()->event(self::LOG_CALLED_GET_TENANT);
        $this->loadOfferingItemsLimitsFromCloud();
        $this->update();
        return $this->toArray();
    }

    /**
     * @return void
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function loadOfferingItemsLimitsFromCloud()
    {
        $this->setLastUserUpdatedTime(time());
        $isForUserOis = $this->memoize(function () {
            $isForUserOis = [];
            $counters = $this->getCountersManager()->getAllCounters(CountersManager::FILTER_EXCLUDE_STORAGE_ADDITIONAL_COUNTERS);
            foreach ($counters as $counter) {
                if ($counter->isForUser()) {
                    $isForUserOis[$counter->getOfferingItemName()] = true;
                }
            }

            return $isForUserOis;
        });

        $this->offeringItems = [];
        $userId = $this->lazyGetPersonalTenantId();

        foreach (ApplicationsManager::AVAILABLE_APPLICATION_TYPES as $applicationType) {
            $application = $this->getCloudApi()->getApplicationByType($applicationType);
            $applicationId = $application->getId();
            $userOfferingItemsFromCloud = $this->getTenantApplicationOfferingItems($userId, $applicationId);
            foreach ($userOfferingItemsFromCloud as $cloudOi) {
                if (!in_array($cloudOi->getName(), $isForUserOis)) {
                    continue;
                }
                $this->offeringItems[] = [
                    'name' => $cloudOi->getName(),
                    'limit' => $cloudOi->hasQuota() && $cloudOi->getQuota()->hasValue() ? $cloudOi->getQuota()->getValue() : -1,
                ];
            }
        }
    }

    /**
     * @return void
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function loadNotificationsFromCloud()
    {
        $userId = $this->lazyGetUserId();
        $user = $this->getCloudApi()->getUser($userId);

        if ($user->hasNotifications() || $user->getNotifications()) {
            $this->notifications = $user->getNotifications();
        } else {
            $this->notifications = self::NOTIFICATIONS_EMPTY;
        }
    }

    /**
     * @verb(GET)
     * @path("/agents")
     * @param(string)
     * @return(object)
     * @throws Exception
     */
    public function getAgents($language)
    {
        if ($this->userstatus != static::USER_STATUS_ACTIVE) {
            return [];
        }

        Log::getInstance()->event(self::LOG_CALLED_GET_AGENTS);
        $mspApi = $this->getMspApi();
        $impersonationToken = $mspApi->getImpersonationUserToken($this->lazyGetUserId());
        $backupConsole = $mspApi->getBackupConsole($this->lazyGetPersonalTenantId(), $language);
        $amsApi = $this->_facade->getAmsApi($backupConsole->getHost());
        $links = $amsApi->getPersonifiedLinks($impersonationToken);
        $agents = $links->getAgents();
        $agentManager = new AgentsManager();

        return $agentManager->buildAgentsListForUser($agents, $language);
    }

    /**
     * @verb(GET)
     * @path("/loginByToken")
     */
    public function loginByToken()
    {
        Log::getInstance()->event(self::LOG_CALLED_GET_WEB_RESTORE_DIGEST);
        $mspApi = $this->getMspApi();
        $brand = $mspApi->getBrand($this->tenant->mspGroupId);
        $url = $brand->getAccountServerUrl();
        $token = $mspApi->getImpersonationUserToken($this->lazyGetUserId());
        return sprintf(self::URL_REDIRECT_MASK, $url, $token);
    }

    /**
     * @verb(GET)
     * @path("/machineUsage")
     * @param()
     */
    public function getMachineUsage()
    {
        Log::getInstance()->event(self::LOG_CALLED_GET_MACHINE_USAGE);
        $backupsItems = $this->_getUserManager()->getBackItems($this->lazyGetPersonalTenantId());

        if (!is_array($backupsItems)) {
            return [];
        }

        $backupList = [];
        foreach ($backupsItems as $item) {
            /** @var Service_AmsApiClient_Response_Struct_BackupRecord $item */
            $backupList[] = [
                self::BACKUP_LIST_NAME => $item->getTitle(),
                self::BACKUP_LIST_STATUS => $item->getStatus(),
                self::BACKUP_LIST_LAST_BACKUP => $this->formatBackupTime($item->getLastBackup()),
                self::BACKUP_LIST_NEXT_BACKUP => $this->formatBackupTime($item->getNextBackup()),
                self::BACKUP_LIST_TYPE => $item->getType()
            ];
        }
        return $backupList;
    }

    /**
     * @verb(GET)
     * @path("/updateStatistics")
     * @param()
     * @access(global,true)
     */
    public function updateStatistics()
    {
        Log::getInstance()->event(self::LOG_CALLED_UPDATE_STATISTICS);

        $usages = [];
        $canManageQuotas = $this->canManageQuotas();
        $userUsages = array_filter($this->getCloudApi()->getTenantUsages($this->lazyGetPersonalTenantId()), function ($usage) {
            return $usage->hasOfferingItem() && $usage->getOfferingItem()->getStatus() === Api::OFFERING_ITEM_STATUS_ACTIVE;
        });
        if (!count($userUsages)) {
            return [];
        }
        $usagesHashmap = Arr::map($userUsages, function ($usage) { return $usage->getName(); });

        $counters = $this->getCountersManager()->findCountersByOfferingItemNames(array_keys($usagesHashmap));
        $offeringItemLimits = Arr::map(
            $usagesHashmap,
            function ($usage) { return $usage->getName(); },
            function ($usage) {
                $offeringItem = $usage->getOfferingItem();
                return $offeringItem->hasQuota() && $offeringItem->getQuota()->hasValue()
                    ? $offeringItem->getQuota()->getValue()
                    : \Model_Resource_Tenant::UNLIMITED_VALUE;
            }
        );

        foreach ($counters as $counter) {
            $offeringItemName = $counter->getOfferingItemName();
            if (!isset($usagesHashmap[$offeringItemName])) {
                continue;
            }

            $counterName = $counter->getName();
            $usages[$counterName] = [
                static::STATISTICS_USAGE => [],
                static::STATISTICS_LIMIT => []
            ];

            $usage = $usagesHashmap[$offeringItemName];
            $usages[$counterName][static::STATISTICS_USAGE] =
                $usage->hasValue() ? $usage->getValue() : 0;

            $tenantLimit = Arr::get($this->tenant, $counterName,
                $canManageQuotas ? 0 : \Model_Resource_Tenant::UNLIMITED_VALUE);
            $offeringItemLimit = $offeringItemLimits[$offeringItemName];
            $usages[$counterName][static::STATISTICS_LIMIT] =
                ($canManageQuotas || $tenantLimit) ? $offeringItemLimit : $tenantLimit;
        }
        return $usages;
    }

    /**
     * @verb(POST)
     * @path("/activateAccount")
     * @param(object,body)
     * @access(admin,true)
     * @access(owner,true)
     * @access(referrer,true)
     */
    public function activateAccount($configuration)
    {
        Log::getInstance()->event(self::LOG_CALLED_ACTIVATE_ACCOUNT);

        $notificationManager = $this->getScreenNotificationManager();
        $notificationManager->activateAccountHasStarted();
        try {
            if ($this->userstatus != self::USER_STATUS_NOT_ACTIVATED) {
                throw new Exception(self::EXCEPTION_USER_ALREADY_ACTIVATED);
            }

            if (!isset($configuration->password)) {
                throw new Exception(self::EXCEPTION_PASSWORD_MUST_BE_PRESENT);
            }

            $this->endUserPassword = $configuration->password;

            $this->_activate();
            $this->refreshEventSubscriptions();

            $this->applyProfileUserQuotaMode($this);
            $this->updatePersonalTenantQuotas();

            $this->update();
        } catch (Exception $e) {
            $notificationManager->activateAccountHasFailed();
            throw $e;
        }
        $notificationManager->activateAccountHasFinished();
    }

    /**
     * @verb(POST)
     * @path("/changePassword")
     * @param(object,body)
     * @access(admin,true)
     * @access(owner,true)
     * @access(referrer,true)
     */
    public function changePassword($configuration)
    {
        Log::getInstance()->event(self::LOG_CALLED_CHANGE_PASSWORD);

        $notificationManager = $this->getScreenNotificationManager();
        $notificationManager->changePasswordHasStarted();
        try {

            if (!isset($configuration->password)) {
                throw new Exception(self::EXCEPTION_PASSWORD_MUST_BE_PRESENT);
            }

            $this->endUserPassword = $configuration->password;
            $this->activateDuringCreation = false; // disable password synchronization

            $userManager = $this->_getUserManager();

            $userManager->resetUserPassword(
                $this->endUserLogin,
                $this->endUserPassword
            );

            $this->update();
        } catch (Exception $e) {
            $notificationManager->changePasswordHasFailed();
            throw $e;
        }
        $notificationManager->changePasswordHasFinished();
    }

    /**
     * @verb(GET)
     * @path("/upgradeUser")
     */
    public function upgradeUser()
    {
        Log::getInstance()->event(sprintf(self::LOG_CALLED_UPGRADE, $this->aps->id));

        if (!$this->offeringItems) {
            try {
                $this->loadOfferingItemsLimitsFromCloud();
            } catch (Exception $e) {
                Log::getInstance()->error("Error getting offering items while upgrade");
                Log::getInstance()->error($e->getMessage());
            }
        }

        if (!$this->notifications) {
            try {
                $this->loadNotificationsFromCloud();
            } catch (Exception $e) {
                $this->notifications = self::NOTIFICATIONS_EMPTY;
                Log::getInstance()->error("Error getting notifications items while upgrade");
                Log::getInstance()->error($e->getMessage());
            }
        }
        $this->update();

        return $this->toArray();
    }

    /* ======= OTHER METHODS ======= */

    public function upgrade()
    {
        /*
        Log::getInstance()->event(sprintf(self::LOG_CALLED_UPGRADE, $this->aps->id));
        $this->convertProperties($this, 'upgrade');

        // todo fix me please !!!
        $this->aps->type = $this->getSelfApsType();
        $this->subscriptionId = $this->tenant->subscriptionId;
        $this->serviceUserId = $this->user->aps->id;
        $this->displayName = $this->user->displayName;
        if (is_null($this->servicestatus)) {
            $this->servicestatus = $this->userstatus == self::USER_STATUS_INACTIVE ?
                self::ACCOUNT_DISABLED :
                self::ACCOUNT_ENABLED;
        }
        if ($this->userstatus == self::USER_STATUS_ACTIVE || $this->userstatus == self::USER_STATUS_INACTIVE) {
            $this->userstatus = $this->user->disabled || $this->servicestatus == self::ACCOUNT_DISABLED ?
                self::USER_STATUS_INACTIVE :
                self::USER_STATUS_ACTIVE;
        }
        if (is_null($this->activateDuringCreation)) {
            $this->activateDuringCreation = true;
        }

        $this->initializeCounters();

        // remove deprecated properties
        if (isset($this->previousStatus)) {
            unset($this->previousStatus);
        }
        if (isset($this->notificationsBackup)) {
            unset($this->notificationsBackup);
        }
        if (isset($this->UserWindowsServerEssentialsCount)) {
            unset($this->UserWindowsServerEssentialsCount); // todo: unset in abc
        }
        if (isset($this->UserVirtualHostsCount)) {
            unset($this->UserVirtualHostsCount);  // todo: unset in abc
        }

        // workaround for empty link to the datacenter
        if ($this->tenant->datacenter->isEmpty()) {
            $datacenterId = $this->tenant->datacenterId;
            if (empty($datacenterId)) {
                throw new Exception(self::EXCEPTION_TENANT_HAS_NOT_LINKED_WITH_DATACENTER_MESSAGE);
            }
            $datacenter = $this->_getDatacenterById($this->tenant->global, $datacenterId);
            $this->tenant->datacenter->set($datacenter);
        }
        // todo fix me please !!!

        $this->update();
        */
    }

    /**
     * Further use in dedicated command which should be called after package upgrade  from .spec
     * @throws Service_BaseClient_HTTPException
     */
    public function syncCloud()
    {
        Log::getInstance()->event(sprintf(self::LOG_CALLED_SYNC, $this->aps->id));

        $this->lazyGetUserId();
        $this->lazyGetPersonalTenantId();
        $this->fillUserLogin();

        $mspApi = $this->getMspApi();
        $userManager = $this->_getUserManager();
        $mspUserInfo = $this->getMspUser(true);
        $version = $mspUserInfo->getVersion();
        $status = $mspUserInfo->getStatus();

        // restore properties from ABC
        $this->endUserLogin = $mspUserInfo->getLogin();
        if (is_null($this->endUserPassword) && $status != Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_INACTIVE) {
            $this->endUserPassword = $this->user->password;
        }
        if (is_null($this->privileged)) {
            $this->privileged = $mspUserInfo->getPrivileged();
        }
        if (is_null($this->notifyBackup)) {
            $this->notifyBackup = $mspUserInfo->getNotificationsBackup();
        }
        if (is_null($this->notificationsManagement)) {
            $this->notificationsManagement = $mspUserInfo->getNotificationsManagement();
        }

        // update properties in ABC
        if ($status == Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_DISABLE) {
            $version = $userManager->enableUser($this->lazyGetUserId(), $version);
        }
        if ($this->user->email != $mspUserInfo->getEmail()) {
            $mspApi->mailUpdate($this->endUserLogin, $this->user->email);
        }
        $updateUserInfo = $this->_getUpdateUserInfo($this);

        $tenant = $this->tenant;
        $userPrivileges = $updateUserInfo->getPrivileges();
        // reset O365 seats to unlimited if it has a limited quota in the previous version
        if (isset($tenant->MailboxCount->limit) && $tenant->MailboxCount->limit) {
            $userPrivileges->setMailboxCount(-1);
        }
        // reset local storage to unlimited if it has a limited quota in the previous version
        if (isset($tenant->LocalStorageUsage->limit) && $tenant->LocalStorageUsage->limit) {
            $userPrivileges->setLocalStorageSize(-1);
        }

        $exception = null;
        try {
            $version = $mspApi->updateUser($this->lazyGetUserId(), $updateUserInfo, $version);
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
            $exception = $e;
        }

        if ($this->userstatus == Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_DISABLE) {
            $userManager->disableUser($this->lazyGetUserId(), $version);
        }

        if (!is_null($exception)) {
            throw $exception;
        }

        // update properties in APS
        $this->update();
    }

    /**
     * @return PartnerCredentials
     */
    public function getPartnerCredentials()
    {
        Log::getInstance()->event(self::LOG_CALLED_GET_PARTNER_CREDENTIALS);
        $datacenter = $this->tenant->datacenter;
        $credentials = new PartnerCredentials();
        $credentials->baasUrl = $datacenter->url;
        $credentials->partnerLogin = $datacenter->login;
        $credentials->partnerPassword = $datacenter->password;
        return $credentials;
    }

    public function lazyGetPersonalTenantId()
    {
        if (!$this->personalTenantId) {
            $this->personalTenantId = $this->resolvePersonalTenantId();
        }

        return $this->personalTenantId;
    }

    /**
     * @return string
     */
    protected function getSelfApsType()
    {
        return self::CLOUD_BACKUP_USERS_TYPE;
    }

    /**
     * @return Model_UserScreenNotificationManagerInterface
     */
    protected function getScreenNotificationManager($user = null)
    {
        if (!isset($this->notificationManager)) {
            $user = is_null($user) ? $this : $user;
            $this->notificationManager = $this->_managerFactory->getUserScreenNotificationManager($user);
        }

        return $this->notificationManager;
    }

    protected function isServiceUserHasSubscription()
    {
        $notificationManager = $this->getScreenNotificationManager();

        $rql = RqlBuilder::createInstance()
            ->implementing(\APS\TypeLibrary::getTypeByServiceId('users'))
            ->linkedWith($this->serviceUserId)
            ->ne('aps.id', $this->aps->id)
            ->limit(0, 1)
            ->getQuery();

        $assignedUsers = $this->_apsc->impersonate($this->tenant)->getResources($rql);
        if (count($assignedUsers) > 0) {
            $notificationManager->provisionHasFailedSubscriptionAlreadyAssigned(
                $this->subscriptionId,
                $assignedUsers[0]->subscriptionId
            );
            return true;
        }
        return false;
    }

    protected function lazyGetUserId()
    {
        if (!$this->userId) {
            if (!$this->mspUserId) {
                // here is situation when data seems to be fully incorrect
                $this->reinitializeNullProperties(['userId', 'mspUserId', 'offeringItems', 'notifications']);
                return $this->userId;
            } else {
                // real lazy loading when user is old (created with mspApi) and not yet loaded
                $user = $this->getMspUser();
                if (is_null($user)) {
                    return null;
                }

                $this->userId = $user->getUuid();
            }
        }

        return $this->userId;
    }

    /**
     * @param array $properties
     * @throws Exception
     */
    protected function reinitializeNullProperties($properties) {
        $user = $this->fetchUserFromAnotherApsInstance();
        if (!$user) {
            Log::getInstance()->notice('Unable to reinitialize (APSc bug w/a) user "%s" properties.', $user->aps->id);
            return;
        }
        foreach ($properties as $property) {
            if ($this->$property !== null) {
                continue;
            }
            $this->$property = $user->$property;
        }
    }

    protected function fillUserLogin()
    {
        if (!$this->login) {
            $user = $this->getMspUser();
            if (is_null($user)) {
                return;
            }
            $this->login = $user->getLogin();
        }
    }

    protected function getNotifications($user)
    {
        $hasNotifications = $user->notifications !== self::NOTIFICATIONS_EMPTY;
        return $hasNotifications ? $user->notifications : [];
    }

    /**
     * @param $user
     * @return array
     */
    protected function getUpdatableNotifications($user)
    {
        $notifications = $this->getNotifications($user);
        return NotificationManager::getUpdatableNotifications($notifications);
    }

    /**
     * @param $user
     * @return array
     */
    protected function getSubscriptiableNotifications($user)
    {
        $notifications = $this->getNotifications($user);
        return NotificationManager::getSubscriptiableNotifications($notifications);
    }

	/**
	 * @param static $user
	 * @param null|array $notifications
	 * @throws Exception
	 */
    protected function updateUserSubscriptiableNotifications($user, $notifications = null)
    {
        try {
            $notifications = is_null($notifications) ? $this->getSubscriptiableNotifications($user) : $notifications;
            $this->getNotificationManager($this->getCloudApi())->updateUserNotifications($user->lazyGetUserId(), $notifications);

        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
        }
    }

    private function fetchUserFromAnotherApsInstance()
    {
        return $this->memoize(function () {
            $apiFactory = new ApiFactory();
            $token = $this->getOssApi()->getProviderToken();
            $apsc = $apiFactory->getApsController(
                $token[OssApiInterface::PROPERTY_CONTROLLER_URI],
                null
            );
            $this->_apsc = $apsc;
            foreach ($this->tenant->users as $user) {
                if ($user->aps->id === $this->aps->id) {
                    return $user;
                }
            }
            return null;
        }, $this->aps->id);
    }

    /**
     * @param bool $force
     * @return Service_MspApiClient_Response_Struct_UserInfo
     */
    private function getMspUser($force = false)
    {
        return $this->memoize(function() {
            $userId = !empty($this->userId) ? $this->userId : $this->mspUserId;
            if (!$userId) {
                return null;
            }
            return $this->getMspApi()->getUserById($userId);
        }, '', $force);
    }

    /**
     * todo remove this w/a after fix of ci-lib-php-cloud-client (array of one element => single string)
     * @param $user
     * @return string
     */
    private function resolvePersonalTenantId($user = null)
    {
        $user = $user ?: $this->getUserInfo($this->lazyGetUserId());

        return $this->_getUserManager()->getUserPersonalTenantId($user);
    }

    /**
     * @param users $resource
     * @param $method
     * @param null $destination
     */
    private function convertProperties($resource, $method)
    {
        $converter = new UsersConverterHelper($method);
        $tenantApplications = $this->getTenantApplications();
        $applicationTypes = array_map(function ($application){
            return $application->getType();
        }, $tenantApplications);
        $converter->setDefaultApplications($applicationTypes);

        if ($method === 'provision') {
            $converter->setDefaultNotifications($this->tenant->profile->notifications ?
                $this->tenant->profile->notifications : self::NOTIFICATIONS_EMPTY
            );
        } elseif ($method === 'upgrade') {
            $converter->setConvertOfferingItems(!is_array($resource->offeringItems));
        }
        $converter->convert($resource, $resource);
    }

    /**
     * @return mixed
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function getTenantApplications()
    {
        $applicationsManager = new ApplicationsManager($this->tenant, $this->getMspApi());

        return $applicationsManager->getTenantApplications($this->tenant->tenantId);
    }

    private function isApplicationEnabledForTenant($applicationType)
    {
        return $this->fromCache(function () use ($applicationType) {
            // todo make getting id of application slightly easier
            $applicationsManager = new ApplicationsManager($this->tenant, $this->getMspApi());
            $backupApplicationId = $applicationsManager->getApplicationIdByType($applicationType);

            return $this->getCloudApi()->isApplicationEnabledForTenant($this->tenant->tenantId, $backupApplicationId);
        }, static::TENANT_APPLICATIONS_CACHE_TTL, false, $applicationType . '_' . $this->tenant->aps->id);
    }

    /**
     * @return bool
     */
    private function canManageQuotas()
    {
        $profile = $this->tenant->profile;
        if ($profile->isEmpty()) {
            $userQuota = true;
        } else {
            $userQuota = $profile->userQuota !== Service_APSApi_Api::PROFILE_USER_QUOTA_OFF;
        }
        return $userQuota;
    }

    private function _activate()
    {
        $userManager = $this->_getUserManager();
        $userInfo = $this->getMspApi()->getUserById($this->lazyGetUserId());
        if ($userInfo->getStatus() == Service_MspApiClient_Response_Struct_UserInfo::STATUS_INACTIVE) {
            $userManager->activateUser(
                $this->endUserLogin,
                $this->endUserPassword
            );
        }


        $this->userstatus = self::USER_STATUS_ACTIVE;
    }

    protected function throwRestAccepted()
    {
        throw new \Rest\Accepted(
            $this,
            sprintf(
                self::LOG_FORMAT_CREATING_USER_TASK,
                $this->endUserLogin,
                $this->subscriptionId
            ),
            self::ASYNC_OPERATION_TIMEOUT
        );
    }

    /**
     * @return Model_UserManager
     */
    private function _getUserManager()
    {
        if (!isset($this->_userManager)) {
            $mspApi = $this->getMspApi();
            $this->_userManager = $this->_managerFactory->getUserManager($mspApi);
        }
        return $this->_userManager;
    }

    /**
     * @param string $time
     * @return int
     */
    private function formatBackupTime($time)
    {
        return $time ? strtotime($time) : null;
    }

    /**
     * @param $new
     * @return Service_MspApiClient_Request_Struct_UserInfo
     * @throws Exception
     */
    private function _getUpdateUserInfo($new)
    {
        $userInfo = new Service_MspApiClient_Request_Struct_UpdateUserInfo();
        $userInfo->setPrivileged($new->privileged);
        $userInfo->setFirstName($new->user->givenName ? $new->user->givenName : " ");
        $userInfo->setLastName($new->user->familyName ? $new->user->familyName : " ");
        $this->setupUserInfoLocale($userInfo);
        return $this->_initUserInfoPrivileges($userInfo, $new);
    }

    /**
     * @param $new
     * @param null $version
     * @param null $updateStatus
     * @return UserPut
     * @throws Exception
     */
    private function _getUpdateCloudUserInfo($new, $version = null, $updateStatus = null)
    {
        $userUpdate = new UserPut();
        $userUpdate->setNotifications($this->getUpdatableNotifications($new));

        if (!is_null($updateStatus)) {
            $userUpdate->setEnabled($updateStatus === self::ACCOUNT_ENABLED);
        }
        $userUpdate->setContact(new Contact());
        $userUpdate->getContact()->setFirstname($new->user->givenName ? $new->user->givenName : " ");
        $userUpdate->getContact()->setLastName($new->user->familyName ? $new->user->familyName : " ");
        $this->setupUserInfoLocale($userUpdate);
        if (is_null($version)) {
            $userInfo = $this->getUserInfo(true);
            $version = $userInfo->getVersion();
        }
        $userUpdate->setVersion($version);

        return $userUpdate;
    }

    /**
     * @param Service_MspApiClient_Request_Struct_UserInfo|Service_MspApiClient_Request_Struct_UpdateUserInfo|UserPut $userInfo
     * @throws Exception
     */
    protected function setupUserInfoLocale($userInfo)
    {
        $user = $this->user;
        if ($user->locale) {
            $userInfo->setLanguage($user->locale);
        } else {
            Log::getInstance()->warning(
                'The locale is not specified for user "%s".',
                $user->id
            );
        }
    }

    /**
     * @param Service_MspApiClient_Request_Struct_UserInfo $userInfo
     * @param users $usersObject
     * @return Service_MspApiClient_Request_Struct_UserInfo
     */
    private function _initUserInfoPrivileges(Service_MspApiClient_Request_Struct_UserInfoInterface $userInfo, $usersObject)
    {
        if ($usersObject->UserWorkstationCount) {
            $userInfo->getPrivileges()->setWorkstationCount($usersObject->UserWorkstationCount->limit);
        }
        if ($usersObject->UserMobileDeviceCount) {
            $userInfo->getPrivileges()->setMobileCount($usersObject->UserMobileDeviceCount->limit);
        }
        if ($usersObject->userServerCount) {
            $userInfo->getPrivileges()->setServerCount($usersObject->userServerCount->limit);
        }
        if ($usersObject->UserVmCount) {
            $userInfo->getPrivileges()->setVmCount($usersObject->UserVmCount->limit);
        }
        if ($usersObject->userStorageUsage) {
            $userInfo->getPrivileges()->setStorageSize($usersObject->userStorageUsage->limit);
        }
        if ($usersObject->userWebHostingServerCount) {
            $userInfo->getPrivileges()->setWebHostingServerCount($usersObject->userWebHostingServerCount->limit);
        }
        if ($usersObject->UserWebsiteCount) {
            $userInfo->getPrivileges()->setWebsiteCount($usersObject->UserWebsiteCount->limit);
        }
        return $userInfo;
    }

    private function isServiceUserAdministrator()
    {
        return !is_null($this->_getApsApi()->getStaffMember($this->tenant, $this->user->aps->id));
    }

    private function initializeCounters()
    {
        foreach (static::getAllCountersNames() as $counter) {
            if (is_null($this->{$counter})) {
                $this->{$counter} = new Counter();
            }
            if (is_null($this->{$counter}->limit)) {
                $this->{$counter}->limit = 0;
            }
            if (is_null($this->{$counter}->usage)) {
                $this->{$counter}->usage = 0;
            }
        }
    }

    /**
     * @return Api
     */
    protected function getCloudApi()
    {
        return $this->memoize(function() {
            return $this->createCloudApiInstance($this->tenant->datacenter);
        });
    }

    /**
     * @param bool $force
     * @return \Acronis\Cloud\Client\Model\Users\User
     */
    protected function getUserInfo($force = false)
    {
        $userUuid = $this->lazyGetUserId();
        return $this->memoize(function () use ($userUuid) {
            return $this->getCloudApi()->getUser($userUuid);
        }, $userUuid, $force);
    }

    /**
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function updatePersonalTenantQuotas()
    {
        // todo uncomment if needed
//        $this->validateSubscriptionPresent();
        // init personal tenant id if not present
        $this->lazyGetPersonalTenantId();
        $this->deleteTenantApplicationOfferingItemsCache();
        foreach (ApplicationsManager::AVAILABLE_APPLICATION_TYPES as $applicationType) {
            $this->_getUserManager()->updateUserPersonalQuotas($this, $applicationType);
        }
    }

    /**
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function validateSubscriptionPresent()
    {
        $offeringItems = array_filter($this->fetchAllOfferingItems(), function ($offeringItem) {
            return $offeringItem->getStatus() === Api::OFFERING_ITEM_STATUS_ACTIVE;
        });
        $applicationOfferingItems = Arr::map(
            $offeringItems,
            function ($offeringItem) { return $offeringItem->getName(); },
            null,
            false
        );

        $hasActiveOfferingItem = false;
        foreach ($this->fetchSubscriptionCounters() as $subscriptionCounter) {
            $offeringItemName = $subscriptionCounter->getOfferingItemName();
            if (isset($applicationOfferingItems[$offeringItemName])) {
                $hasActiveOfferingItem = true;
                break;
            }
        }

        if (!$hasActiveOfferingItem) {
            $tenant = new \tenant();
            $tenant->fillResource($this->tenant->aps->id, true);
            $tenant->onSubscriptionChange(null);
        }
    }

    /**
     * @return OfferingItemOutput[]
     * @throws RestException
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function fetchAllOfferingItems()
    {
        $cloudApi = $this->getCloudApi();
        $result = [];

        foreach (ApplicationsManager::AVAILABLE_APPLICATION_TYPES as $applicationType) {
            $application = $cloudApi->getApplicationByType($applicationType);
            if (is_null($application)) {
                throw new RestException(404, Str::format(
                    'Cannot fetch application "%s".',
                    $this->type
                ));
            }
            $applicationId = $application->getId();

            $result = array_merge($result, $this->getCloudApi()->getTenantApplicationOfferingItems($this->tenant->tenantId, $applicationId));
        }

        return $result;
    }

    /**
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    protected function updateUserRoles()
    {
        $accessPolicies = array_map(function ($role) {
            $accessPolicy = new AccessPolicy();
            $accessPolicy->setId(Api::NULL_UUID);
            $accessPolicy->setIssuerId(Api::NULL_UUID);
            $accessPolicy->setTrusteeId($this->lazyGetUserId());
            $accessPolicy->setTrusteeType(AccessPolicyTrusteeType::USER);
            $accessPolicy->setVersion(0);
            $accessPolicy->setTenantId($this->tenant->tenantId);
            $accessPolicy->setRoleId($role);

            return $accessPolicy;
        }, $this->serviceRoles);

        $this->getCloudApi()->updateUserAccessPolicies($this->lazyGetUserId(), $accessPolicies);
    }

    /**
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    public function obtainNewServiceRoles()
    {
        if (!$this->isServiceUserAdministrator()) {

            foreach (ApplicationsManager::AVAILABLE_APPLICATION_TYPES as $applicationType) {
                $applicationUserRole = self::APPLICATION_USER_ROLE_MAP[$applicationType] ?? '';
                if ($applicationUserRole && $this->isApplicationEnabledForTenant($applicationType) && !$this->isUserHasServiceRole($applicationUserRole)) {
                    array_push($this->serviceRoles, $applicationUserRole);
                }
            }

            $this->updateUserRoles();
        }
    }

    /**
     * @param $role
     * @return bool
     */
    private function isUserHasServiceRole($role)
    {
        return (bool)Arr::find($this->serviceRoles, function ($userRole) use ($role) {
            return $role == $userRole;
        });
    }

    private function createCloudUser()
    {
        try {
            $user = $this->createUser();
        } catch (Exception $e) {
            if ($e->getCode() != Service_BaseClient_RestTransport::HTTP_CODE_UNKNOWN_STATUS) {
                throw new Exception($e->getMessage());
            }

            $mspGroupId = $this->tenant->mspGroupId;
            $existingUser = $this->findUserByLogin($mspGroupId, $this->endUserLogin);
            if (empty($existingUser)) {
                throw new Exception(self::EXCEPTION_USER_ALREADY_EXISTS);
            }
            $this->mspUserId = $existingUser->getId();
            $mspUser = $this->getMspUser();
            $this->userId = $mspUser->getUuid();
            $this->personalTenantId = $this->resolvePersonalTenantId();
            $cloudApi = $this->getCloudApi();
            $adminUser = $this->_getApsApi()->getAdmin($this->tenant);
            if (is_null($adminUser)) {
                $this->initUserFormProfile($this);
                $cloudApi->updateUser($this->userId, $this->_getUpdateCloudUserInfo($this));
				$this->updateUserSubscriptiableNotifications($this);
            } else {
                $userNotifications = $cloudApi->getUser($this->userId)->getNotifications();
                if ($userNotifications != $this->getUpdatableNotifications($this)) {
                    $cloudApi->updateUser($this->userId, $this->_getUpdateCloudUserInfo($this));
                }
				$this->updateUserSubscriptiableNotifications($this);
                $isUnlimited = $this->hasUnlimitedUserQuota();
                if ($isUnlimited && !count($cloudApi->getOfferingItemsAvailableForChildHashMap($this->personalTenantId))) {
                    foreach ($this->offeringItems as $offeringItem) {
                        $offeringItem->limit = \Model_Resource_Tenant::UNLIMITED_VALUE;
                    }
                }
            }
            return;
        }

		$this->userId = $user->getId();
		$mspUser = $this->getMspApi()->getUserById($user->getId());
		$this->mspUserId = $mspUser->getId();
		$ids = $user->getPersonalTenantId();
		$this->personalTenantId = reset($ids);
		$this->updateUserSubscriptiableNotifications($this);
    }

    /**
     * @return \Acronis\Cloud\Client\Model\Users\User
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function createUser()
    {
        $userBody = new UserPost();
        $userBody->setTenantId($this->tenant->tenantId);
        $userBody->setLogin($this->endUserLogin);

        $contact = new Contact();
        if ($this->user->givenName) {
            $contact->setFirstname($this->user->givenName);
        }
        if ($this->user->familyName) {
            $contact->setLastName($this->user->familyName);
        }
        $contact->setEmail($this->user->email ?: $this->endUserLogin);

        $userBody->setContact($contact);

        $userBody->setNotifications($this->getUpdatableNotifications($this));

        return $this->getCloudApi()->createUser($userBody);
    }

    private function updateMspUser()
    {
        Log::getInstance()->event(self::LOG_MIGRATION_WAS_STARTED);

        $mspApi = $this->getMspApi();
        $userInfo = $this->getMspUser(true);
        $this->endUserLogin = $userInfo->getLogin();
        if ($this->user->email != $userInfo->getEmail()) {
            $mspApi->mailUpdate($this->endUserLogin, $this->user->email);
        }
        $updateUserInfo = $this->_getUpdateUserInfo($this);
        $this->getMspApi()->updateUser($this->lazyGetUserId(), $updateUserInfo, $userInfo->getVersion());
    }

    private function applyProfileUserQuotaMode($user)
    {
        if (!$this->canManageQuotas()) {
            if (empty($user->offeringItems)) {
                // old format, used in mspApi
                $user->userStorageUsage->limit = Service_MspApiClient_Request_Struct_UserPrivileges::USER_UNLIMITED_VALUE;
                $user->userServerCount->limit = Service_MspApiClient_Request_Struct_UserPrivileges::USER_UNLIMITED_VALUE;
                $user->UserVmCount->limit = Service_MspApiClient_Request_Struct_UserPrivileges::USER_UNLIMITED_VALUE;
                $user->UserWorkstationCount->limit = Service_MspApiClient_Request_Struct_UserPrivileges::USER_UNLIMITED_VALUE;
                $user->UserMobileDeviceCount->limit = Service_MspApiClient_Request_Struct_UserPrivileges::USER_UNLIMITED_VALUE;
                $user->userWebHostingServerCount->limit = Service_MspApiClient_Request_Struct_UserPrivileges::USER_UNLIMITED_VALUE;
                $user->UserWebsiteCount->limit = Service_MspApiClient_Request_Struct_UserPrivileges::USER_UNLIMITED_VALUE;
            } else {
                foreach ($user->offeringItems as $index => $offeringItem) {
                    $user->offeringItems[$index]->limit = Service_MspApiClient_Request_Struct_UserPrivileges::USER_UNLIMITED_VALUE;
                }
            }
        }
    }

    private function _activateDuringCreation()
    {
        if (!$this->activateDuringCreation) {
            $this->userstatus = self::USER_STATUS_NOT_ACTIVATED;
            return;
        }
        $this->_activate();
    }

    /**
     * @param $mspGroupId
     * @param $endUserLogin
     * @return null|Service_MspApiClient_Response_Struct_UserItem
     */
    private function findUserByLogin($mspGroupId, $endUserLogin)
    {
        $users = $this->getMspApi()->getUsers($mspGroupId);
        foreach ($users as $user) {
            if ($user->getLogin() === $endUserLogin) {
                return $user;
            }
        }
        return null;
    }

    protected static function getAllCountersNames()
    {
        return array_values(Model_Resource_Tenant::$userQuotasPropertiesMap);
    }

    /**
     * @param string $tenantId
     * @param string $applicationId
     * @return string
     */
    private function getCacheKeyForTenantOfferingItemLimits($tenantId, $applicationId = '')
    {
        return sprintf(self::CACHE_KEY_FOR_APPLICATION_OI, $tenantId, $applicationId);
    }

    /**
     * @return void
     */
    private function deleteTenantApplicationOfferingItemsCache()
    {
        $cache = $this->getCache();
        $userId = $this->lazyGetPersonalTenantId();
        $cacheKey = $this->getCacheKeyForTenantOfferingItemLimits($userId);
        $cache->deleteAll($cacheKey);
    }

    /**
     * @param string $tenantId
     * @param string $applicationId
     * @return OfferingItemOutput[]|mixed
     * @throws \Acronis\Cloud\Client\ApiException
     * @throws \Acronis\Cloud\Client\HttpException
     */
    private function getTenantApplicationOfferingItems($tenantId, $applicationId)
    {
        $cache = $this->getCache();
        $cacheKey = $this->getCacheKeyForTenantOfferingItemLimits($tenantId, $applicationId);

        if ($cache->has($cacheKey)) {
            $userOfferingItemsFromCloud = $cache->get($cacheKey);
        } else {
            $userOfferingItemsFromCloud = $this->getCloudApi()->getTenantApplicationOfferingItems($tenantId, $applicationId);
            $cache->set($cacheKey, $userOfferingItemsFromCloud);
        }

        return $userOfferingItemsFromCloud;
    }
}

